from __future__ import annotations

import copy
import functools
import inspect
import uuid
from collections.abc import Callable, Iterable
from contextlib import contextmanager
from functools import partial
from typing import TYPE_CHECKING, Any, overload

import sqlalchemy as sa

from pydiverse.pipedag._typing import CallableT
from pydiverse.pipedag.container import Blob, RawSql, Table
from pydiverse.pipedag.context import ConfigContext, RunContext, TaskContext
from pydiverse.pipedag.context.context import (
    CacheValidationMode,
)
from pydiverse.pipedag.core.task import Task, TaskGetItem, UnboundTask
from pydiverse.pipedag.errors import CacheError
from pydiverse.pipedag.materialize.cache import TaskCacheInfo, task_cache_key
from pydiverse.pipedag.util import deep_map
from pydiverse.pipedag.util.computation_tracing import (
    ComputationTraceRef,
    ComputationTracerProxy,
)
from pydiverse.pipedag.util.hashing import stable_hash
from pydiverse.pipedag.util.json import PipedagJSONEncoder

if TYPE_CHECKING:
    from pydiverse.pipedag import Flow, Stage
    from pydiverse.pipedag.backend.table.base import TableHook
    from pydiverse.pipedag.materialize.store import PipeDAGStore


@overload
def materialize(
    *,
    name: str | None = None,
    input_type: type | None = None,
    version: str | None = None,
    cache: Callable[..., Any] | None = None,
    lazy: bool = False,
    nout: int = 1,
    add_input_source: bool = False,
    ordering_barrier: bool | dict[str, Any] = False,
    call_context: Callable[[], Any] | None = None,
) -> Callable[[CallableT], CallableT | UnboundMaterializingTask]:
    ...


@overload
def materialize(fn: CallableT, /) -> CallableT | UnboundMaterializingTask:
    ...


def materialize(
    fn: CallableT | None = None,
    *,
    name: str | None = None,
    input_type: type | None = None,
    version: str | None = None,
    cache: Callable[..., Any] | None = None,
    lazy: bool = False,
    group_node_tag: str | None = None,
    nout: int = 1,
    add_input_source: bool = False,
    ordering_barrier: bool | dict[str, Any] = False,
    call_context: Callable[[], Any] | None = None,
):
    """Decorator to create a task whose outputs get materialized.

    This decorator takes a class and turns it into a :py:class:`MaterializingTask`.
    This means, that this function can only be used as part of a flow.
    Any outputs it produces get written to their appropriate storage backends
    (either the table or blob store). Additionally, any :py:class:`Table`
    or :py:class:`Blob` objects this task receives as an input get replaced
    with the appropriate object retrieved from the storage backend.
    In other words: All outputs from a materializing task get written to the store,
    and all inputs get retrieved from the store.

    Because of how caching is implemented, task inputs and outputs must all
    be "materializable". This means that they can only contain objects of
    the following types:
    ``dict``, ``list``, ``tuple``,
    ``int``, ``float``, ``str``, ``bool``, ``None``,
    ``datetime.date``, ``datetime.datetime``, ``pathlib.Path``,
    :py:class:`Table`, :py:class:`RawSql`, :py:class:`Blob`, or :py:class:`Stage`.

    :param fn:
        The function that gets executed by this task.
    :param name:
        The name of this task.
        If no `name` is provided, the name of `fn` is used instead.
    :param input_type:
        The data type as which to retrieve table objects from the store.
        All tables passed to this task get loaded from the table store and converted
        to this type. See :doc:`Table Backends </table_backends>` for
        more information.
    :param version:
        The version of this task.
        Unless the task is lazy, you always need to manually change this
        version number when you change the implementation to ensure that the
        task gets executed and the cache flushed.

        If the `version` is ``None`` and the task isn't lazy, then the task always
        gets executed, and all downstream tasks get invalidated.
    :param cache:
        An explicit cache function used to determine cache validity of the task inputs.

        This function gets called every time before the task gets executed.
        It gets called with the same arguments as the task.

        An explicit function for validating cache validity. If the output
        of this function changes while the source parameters are the same (e.g.
        the source is a filepath and `cache` loads data from this file), then the
        cache will be deemed invalid and is not used.
    :param lazy:
        Whether this task is lazy or not.

        Unlike a normal task, lazy tasks always get executed. However, if a lazy
        task produces a lazy table (e.g. a SQL query), the table store checks if
        the same query has been executed before. If this is the case, then the
        query doesn't get executed, and instead, the table gets copied from the cache.

        This behaviour is very useful, because you don't need to manually bump
        the `version` of a lazy task. This only works because for lazy tables
        generating the query is very cheap compared to executing it.
    :param group_node_tag:
        Set a tag that may add this task to a configuration based group node.
    :param nout:
        The number of objects returned by the task.
        If set, this allows unpacking and iterating over the results from the task.
    :param add_input_source:
        If true, Table and Blob objects are provided as tuple together with the
        dematerialized object. ``task(a=tbl)`` will result in
        ``a=(dematerialized_tbl, tbl)`` in the task.
    :param ordering_barrier:
        If true, the task will be surrounded by a GroupNode(ordering_barrier=True).
        If a dictionary is provided, it is surrounded by a
        GroupNode(**ordering_barrier). This allows passing style, style_tag, and label
        arguments.
    :param call_context:
        An optional context manager function that is opened before the task or its
        optional cache function is called and closed afterward.

    Example
    -------

    ::

        @materialize(version="1.0", input_type=pd.DataFrame)
        def task(df: pd.DataFrame) -> pd.DataFrame:
            df = ...  # Do something with the dataframe
            return Table(df)

        @materialize(lazy=True, input_type=sa.Table)
        def lazy_task(tbl: sa.Alias) -> sa.Table:
            query = sa.select(tbl).where(...)
            return Table(query)

    You can use the `cache` argument to specify an explicit cache function.
    In this example we define a task that reads a dataframe from a csv file.
    Without specifying a cache function, this task would only get rerun when the
    `path` argument changes. Instead, we define a function that calculates a
    hash based on the contents of the csv file, and pass it to the `cache` argument
    of ``@materialize``. This means that the task now gets invalidated and rerun if
    either the `path` argument changes or if the content of the file pointed
    to by `path` changes::

        import hashlib

        def file_digest(path):
            with open(path, "rb") as f:
                return hashlib.file_digest(f, "sha256").hexdigest()

        @materialize(version="1.0", cache=file_digest)
        def task(path):
            df = pd.read_scv(path)
            return Table(df)

    Setting nout allows you to use unpacking assignment with the task result::

        @materialize(version="1.0", nout=2)
        def task():
            return 1, 2

        # Later, while defining a flow, you can use unpacking assignment
        # because you specified that the task returns 2 objects (nout=2).
        one, two = task()

    If a task returns a dictionary or a list, you can use square brackets to
    explicitly wire individual values / elements to task inputs::

        @materialize(version="1.0")
        def task():
            return {
                "x": [0, 1],
                "y": [2, 3],
            }

        # Later, while defining a flow
        # This would result in another_task being called with ([0, 1], 3).
        task_out = task()
        another_task(task_out["x"], task_out["y"][1])
    """
    if fn is None:
        return partial(
            materialize,
            name=name,
            input_type=input_type,
            version=version,
            cache=cache,
            lazy=lazy,
            group_node_tag=group_node_tag,
            nout=nout,
            add_input_source=add_input_source,
            ordering_barrier=ordering_barrier,
            call_context=call_context,
        )

    return UnboundMaterializingTask(
        fn,
        name=name,
        input_type=input_type,
        version=version,
        cache=cache,
        lazy=lazy,
        group_node_tag=group_node_tag,
        nout=nout,
        add_input_source=add_input_source,
        ordering_barrier=ordering_barrier,
        call_context=call_context,
    )


class UnboundMaterializingTask(UnboundTask):
    """A materializing task without any bound arguments.

    Instances of this class get initialized using the
    :py:func:`@materialize <pydiverse.pipedag.materialize>` decorator.
    By calling this object inside a ``with Flow(...)`` context,
    a :py:class:`~.MaterializingTask` gets created that binds the provided
    arguments to the task.

    >>> @materialize
    ... def some_task(arg):
    ...     ...
    ...
    >>> type(some_task)
    <class 'pydiverse.pipedag.materialize.core.UnboundMaterializingTask'>

    >>> with Flow() as f:
    ...     with Stage("stage"):
    ...         x = some_task(...)
    ...
    >>> type(x)
    <class 'pydiverse.pipedag.materialize.core.MaterializingTask'>

    When you call a `UnboundMaterializingTask` outside a flow definition context,
    then the original function (that was decorated with :py:func:`@materialize
    <pydiverse.pipedag.materialize>`) gets called. The only difference being, that
    any :py:class:`~.Table`, :py:class:`~.Blob` or :py:class:`~.RawSql` objects
    returned from the task get replaced with the object that they wrap:

    >>> import pandas as pd
    >>>
    >>> @materialize(input_type=pd.DataFrame)
    ... def double_it(df):
    ...     return Table(df * 2)
    ...
    >>> x = pd.DataFrame({"x": [0, 1, 2, 3]})
    >>> double_it(x)
       x
    0  0
    1  2
    2  4
    3  6


    """

    def __init__(
        self,
        fn: Callable,
        *,
        name: str | None = None,
        input_type: type | None = None,
        version: str | None = None,
        cache: Callable[..., Any] | None = None,
        lazy: bool = False,
        group_node_tag: str | None = None,
        nout: int = 1,
        add_input_source: bool = False,
        ordering_barrier: bool | dict[str, Any] = False,
        call_context: Callable[[], Any] | None = None,
    ):
        super().__init__(
            MaterializationWrapper(fn),
            name=name,
            nout=nout,
        )

        self._bound_task_type = MaterializingTask
        self._original_fn = fn

        self.input_type = input_type
        self.version = version
        self.cache = cache
        self.lazy = lazy
        self.group_node_tag = group_node_tag
        self.add_input_source = add_input_source
        self.call_context = call_context

        if isinstance(ordering_barrier, bool):
            group_node_args = {"ordering_barrier": ordering_barrier}
        elif isinstance(ordering_barrier, dict):
            group_node_args = ordering_barrier.copy()
            if "ordering_barrier" not in group_node_args:
                group_node_args["ordering_barrier"] = True

            group_node_keys = {"ordering_barrier", "label", "style", "style_tag"}
            unexpected_keys = set(ordering_barrier.keys()) - group_node_keys
            if unexpected_keys:
                raise ValueError(
                    "If ordering_barrier is a dictionary, it must only contain keys "
                    "that are valid parameters of class GroupNode constructor."
                    f"{unexpected_keys} are unexpected."
                )
        else:
            raise ValueError(
                "The ordering_barrier parameter must be a boolean or a dictionary."
            )
        self.group_node_args = group_node_args

        if version is AUTO_VERSION:
            if input_type is None:
                # The input_type is used to determine which type of auto versioning
                # to use (either not supported, lazy or tracing)
                raise ValueError("Auto-versioning task must specify an input type")
            if lazy:
                raise ValueError(
                    "Task can't be lazy and auto-versioning at the same time"
                )

    def __call__(self, *args, **kwargs) -> MaterializingTask:
        if self.group_node_args["ordering_barrier"]:
            from pydiverse.pipedag import GroupNode

            with GroupNode(**self.group_node_args):
                return super().__call__(*args, **kwargs)  # type: ignore
        return super().__call__(*args, **kwargs)  # type: ignore

    def _call_original_function(self, *args, **kwargs):
        try:
            task_context = TaskContext.get()  # this may raise Lookup Error

            # this is a subtask call. Thus we demand identical input_types
            sub_task: MaterializingTask = task_context.task  # type: ignore
            if (
                sub_task.input_type is not None
                and self.input_type is not None
                and sub_task.input_type != self.input_type
            ):
                raise RuntimeError(
                    f"Subtask input type {self.input_type} does not match parent task "
                    f"input type {sub_task.input_type}: "
                    f"task={self}, sub_task={sub_task}"
                )
        except LookupError:
            # this is expected when calling the task outside of flow declaration
            pass
        result = self._original_fn(*args, **kwargs)

        def unwrap_mutator(x):
            if isinstance(x, (Table, Blob)):
                return x.obj
            if isinstance(x, RawSql):
                return x.sql
            return x

        return deep_map(result, unwrap_mutator)


class MaterializingTask(Task):
    """
    A pipedag task that materializes all its outputs.

    Instances of this class get initialized by calling a
    :py:class:`~.UnboundMaterializingTask`.

    As a user, you will encounter `MaterializingTask` objects, together with
    :py:class:`~.MaterializingTaskGetItem` objects, mostly during flow declaration,
    because they are used to bind the output (or parts of the output) from one
    task, to the input of another task.
    """

    def __init__(
        self,
        unbound_task: UnboundMaterializingTask,
        bound_args: inspect.BoundArguments,
        flow: Flow,
        stage: Stage,
    ):
        super().__init__(unbound_task, bound_args, flow, stage)

        self.input_type = unbound_task.input_type
        self.add_input_source = unbound_task.add_input_source
        self._version = unbound_task.version
        self.cache = unbound_task.cache
        self.lazy = unbound_task.lazy

    @property
    def version(self):
        try:
            if v := TaskContext.get().override_version:
                return v
        except LookupError:
            pass

        return self._version

    @version.setter
    def version(self, version):
        try:
            TaskContext.get().override_version = version
        except LookupError:
            raise AttributeError(
                "Can't set task version outside of a TaskContext"
            ) from None

    def __getitem__(self, item) -> MaterializingTaskGetItem:
        """Construct a :py:class:`~.MaterializingTaskGetItem`.

        If the corresponding task returns an object that supports
        :external+python:py:meth:`__getitem__ <object.__getitem__>`,
        then this allows you to forward only parts of the task's output to
        the next task.

        Example
        -------
        ::

            @materialize
            def dict_task():
                return {"x": 1, "y": 2}

            @materialize
            def assert_equal(actual, expected):
                assert actual == expected

            with Flow():
                with Stage("stage"):
                    d = dict_task()

                    # Only pass parts of the dictionary returned by dict_task
                    # to the assert_equal tasks.
                    assert_equal(d["x"], 1)
                    assert_equal(d["y"], 2)

        """
        return MaterializingTaskGetItem(self, self, item)

    def run(
        self, inputs: dict[int, Any], ignore_position_hashes: bool = False, **kwargs
    ):
        # When running only a subset of an entire flow, not all inputs
        # get calculated during flow execution. As a consequence, we must load
        # those inputs from the cache.

        def replace_stage_with_pseudo_stage(x):
            if isinstance(x, (Table, Blob, RawSql)):
                x.stage = PseudoStage(x.stage, did_commit=True)
            return x

        for in_id, in_task in self.input_tasks.items():
            if in_id in inputs:
                continue

            store = ConfigContext.get().store
            cached_output, metadata = store.retrieve_most_recent_task_output_from_cache(
                in_task, ignore_position_hashes
            )

            cached_output = deep_map(cached_output, replace_stage_with_pseudo_stage)
            inputs[in_id] = cached_output

        return super().run(inputs, **kwargs)

    def get_output_from_store(
        self, as_type: type = None, ignore_position_hashes: bool = False
    ) -> Any:
        """Retrieves the output of the task from the cache.

        No guarantees are made regarding whether the returned values are still
        up-to-date and cache valid.

        :param as_type: The type as which tables produced by this task should
            be dematerialized. If no type is specified, the input type of
            the task is used.
        :param ignore_position_hashes:
            If ``True``, the position hashes of tasks are not checked
            when retrieving the inputs of a task from the cache.
            This simplifies execution of subgraphs if you don't care whether inputs to
            that subgraph are cache invalid. This allows multiple modifications in the
            Graph before the next run updating the cache.
            Attention: This may break automatic cache invalidation.
            And for this to work, any task producing an input
            for the chosen subgraph may never be used more
            than once per stage.
        :return: The output of the task.
        :raise CacheError: if no outputs for this task could be found in the store.

        Example
        -------

        ::

            # Define some flow
            with Flow() as f:
                with Stage("stage_1"):
                    x = some_task()
                    ...

            # Get output BEFORE calling flow.run()
            df_x = x.get_output_from_store(as_type=pd.DataFrame)

        """
        return _get_output_from_store(
            self, as_type, ignore_position_hashes=ignore_position_hashes
        )


class MaterializingTaskGetItem(TaskGetItem):
    """Object that represents a subset of a :py:class:`~.MaterializingTask` output.

    Instances of this class get initialized by calling
    :py:class:`MaterializingTask.__getitem__`.
    """

    def __init__(
        self,
        task: MaterializingTask,
        parent: MaterializingTask | MaterializingTaskGetItem,
        item: Any,
    ):
        super().__init__(task, parent, item)

    def __getitem__(self, item) -> MaterializingTaskGetItem:
        """
        Same as :py:meth:`MaterializingTask.__getitem__`,
        except that it allows you to further refine the selection.
        """
        return super().__getitem__(item)

    def get_output_from_store(
        self, as_type: type = None, ignore_position_hashes: bool = False
    ) -> Any:
        """
        Same as :py:meth:`MaterializingTask.get_output_from_store()`,
        except that it only loads the required subset of the output.
        """
        return _get_output_from_store(
            self, as_type, ignore_position_hashes=ignore_position_hashes
        )


class MaterializationWrapper:
    """Function wrapper that contains all high level materialization logic

    :param fn: The function to wrap
    """

    def __init__(self, fn: Callable):
        functools.update_wrapper(self, fn)

        self.fn = fn
        self.fn_signature = inspect.signature(fn)

    def __call__(self, *args, **kwargs):
        """Function wrapper / materialization logic

        :param args: The arguments passed to the function
        :param _pipedag_task_: The `MaterializingTask` instance which called
            this wrapper.
        :param kwargs: The keyword arguments passed to the function
        :return: A copy of what the original function returns annotated
            with some additional metadata.
        """

        task_context = TaskContext.get()
        config_context = ConfigContext.get()
        run_context = RunContext.get()

        task: MaterializingTask = task_context.task  # type: ignore
        bound = self.fn_signature.bind(*args, **kwargs)

        if task is None:
            raise TypeError("Task can't be None.")

        # If this is the first task in this stage to be executed, ensure that
        # the stage has been initialized and locked.
        store = config_context.store
        store.ensure_stage_is_ready(task.stage)

        # Compute the cache key for the task inputs
        input_json = store.json_encode(bound.arguments)
        input_hash = stable_hash("INPUT", input_json)

        cache_fn_hash = ""
        if task.cache is not None:
            if config_context.cache_validation.disable_cache_function:
                # choose random cache_fn_hash to ensure downstream tasks are
                # invalidated for FORCE_FRESH_INPUT
                cache_fn_hash = stable_hash("CACHE_FN", uuid.uuid4().hex)
            else:
                cache_fn_output = store.json_encode(task.cache(*args, **kwargs))
                cache_fn_hash = stable_hash("CACHE_FN", cache_fn_output)

        memo_cache_key = task_cache_key(task, input_hash, cache_fn_hash)

        # Check if this task has already been run with the same inputs
        # If yes, return memoized result. This prevents DuplicateNameExceptions
        with run_context.task_memo(task, memo_cache_key) as (success, memo):
            if success:
                task.logger.info(
                    "Task has already been run with the same inputs."
                    " Using memoized results."
                )

                return memo

            # Cache Lookup (only required if task isn't lazy)
            force_task_execution = (
                config_context.cache_validation.mode
                == CacheValidationMode.FORCE_CACHE_INVALID
                or (
                    config_context.cache_validation.mode
                    == CacheValidationMode.FORCE_FRESH_INPUT
                    and task.cache is not None
                )
            )
            skip_cache_lookup = (
                config_context.cache_validation.ignore_task_version
                and task.version != AUTO_VERSION
            ) or task.version is None
            assert_no_fresh_input = (
                config_context.cache_validation.mode
                == CacheValidationMode.ASSERT_NO_FRESH_INPUT
            )

            if task.version is AUTO_VERSION:
                if (
                    force_task_execution
                    and config_context.cache_validation.disable_cache_function
                ):
                    task.version = None
                    skip_cache_lookup = True
                    task.logger.info(
                        "Skipping determination of auto version of task due to forced "
                        "execution and disable_cache_function=True"
                    )
                else:
                    assert not task.lazy

                    auto_version = self.compute_auto_version(task, store, bound)
                    task.version = auto_version
                    task.logger.info(
                        "Assigned auto version to task", version=auto_version
                    )

            if not task.lazy and not skip_cache_lookup and not force_task_execution:
                try:
                    cached_output, cache_metadata = store.retrieve_cached_output(
                        task, input_hash, cache_fn_hash
                    )
                    store.copy_cached_output_to_transaction_stage(
                        cached_output, cache_metadata, task
                    )
                    run_context.store_task_memo(task, memo_cache_key, cached_output)
                    task.logger.info(
                        "Found task in cache. Using cached result.",
                        input_hash=input_hash,
                        cache_fn_hash=cache_fn_hash,
                    )
                    TaskContext.get().is_cache_valid = True
                    run_context.trace_hook.task_cache_status(
                        task, input_hash, cache_fn_hash, cache_metadata, cached_output
                    )
                    return cached_output
                except CacheError as e:
                    task.logger.info(
                        "Failed to retrieve task from cache",
                        input_hash=input_hash,
                        cache_fn_hash=cache_fn_hash,
                        version=task.version,
                        cause=str(e),
                    )
                    TaskContext.get().is_cache_valid = False
                    run_context.trace_hook.task_cache_status(
                        task, input_hash, cache_fn_hash, cache_valid=False
                    )

            if not task.lazy:
                if assert_no_fresh_input and task.cache is not None:
                    raise AssertionError(
                        "cache_validation.mode=ASSERT_NO_FRESH_INPUT is a "
                        "protection mechanism to prevent execution of "
                        "source tasks to keep pipeline input stable. However,"
                        "this task was still triggered."
                    ) from None

            if task.lazy:
                # For lazy tasks, is_cache_valid gets set to false during the
                # store.materialize_task procedure
                TaskContext.get().is_cache_valid = True

                if (
                    config_context.cache_validation.mode
                    == CacheValidationMode.IGNORE_FRESH_INPUT
                ):
                    # `cache_fn_hash` is not used for cache retrieval if
                    # ignore_cache_function is set to True.
                    # In that case, cache_metadata.cache_fn_hash may be different
                    # from the cache_fn_hash of the current task run.

                    try:
                        _, cache_metadata = store.retrieve_cached_output(
                            task, input_hash, ""
                        )
                        cache_fn_hash = cache_metadata.cache_fn_hash
                        run_context.trace_hook.task_cache_status(
                            task, input_hash, cache_fn_hash, cache_metadata, lazy=True
                        )
                    except CacheError as e:
                        task.logger.info(
                            "Failed to retrieve lazy task from cache",
                            cause=str(e),
                            input_hash=input_hash,
                            version=task.version,
                            cache_fn_hash=cache_fn_hash,
                        )

            # Prepare TaskCacheInfo
            if not task.lazy and skip_cache_lookup:
                # Assign random version to disable caching for this task
                task = copy.copy(task)
                task.version = uuid.uuid4().hex

            task_cache_info = TaskCacheInfo(
                task=task,
                input_hash=input_hash,
                cache_fn_hash=cache_fn_hash,
                cache_key=task_cache_key(task, input_hash, cache_fn_hash),
                assert_no_materialization=assert_no_fresh_input,
                force_task_execution=force_task_execution,
            )

            # Compute the input_tables value of the TaskContext
            input_tables = []

            def _input_tables_visitor(x):
                if isinstance(x, Table):
                    input_tables.append(x)
                return x

            deep_map(bound.arguments, _input_tables_visitor)
            task_context.input_tables = input_tables

            # Not found in cache / lazy -> Evaluate Function
            args, kwargs = store.dematerialize_task_inputs(
                task, bound.args, bound.kwargs
            )

            def imperative_materialize(
                table: Table,
                config_context: ConfigContext | None,
                return_as_type: type | None = None,
                return_nothing: bool = False,
            ):
                my_store = config_context.store if config_context is not None else store
                state = task_cache_info.imperative_materialization_state
                if id(table) in state.table_ids:
                    raise RuntimeError(
                        "The table has already been imperatively materialized."
                    )
                table.assumed_dependencies = (
                    list(sorted(state.assumed_dependencies))
                    if len(state.assumed_dependencies) > 0
                    else []
                )
                _ = my_store.materialize_task(
                    task, task_cache_info, table, disable_task_finalization=True
                )
                if not return_nothing:

                    def get_return_obj(return_as_type):
                        if return_as_type is None:
                            return_as_type = task.input_type
                            if (
                                return_as_type is None
                                or not my_store.table_store.get_r_table_hook(
                                    return_as_type
                                ).retrieve_as_reference(return_as_type)
                            ):
                                # dematerialize as sa.Table if it would transfer all
                                # rows to python when dematerializing with input_type
                                return_as_type = sa.Table
                        obj = my_store.dematerialize_item(
                            table, return_as_type, run_context
                        )
                        state.add_table_lookup(obj, table)
                        return obj

                    if isinstance(return_as_type, Iterable):
                        return tuple(get_return_obj(t) for t in return_as_type)
                    else:
                        return get_return_obj(return_as_type)

            task_context.imperative_materialize_callback = imperative_materialize
            run_context.trace_hook.task_pre_call(task)
            result = self.fn(*args, **kwargs)
            run_context.trace_hook.task_post_call(task)
            task_context.imperative_materialize_callback = None
            if task.debug_tainted:
                raise RuntimeError(
                    f"The task {task.name} has been tainted by interactive debugging."
                    f" Aborting."
                )

            def result_finalization_mutator(x):
                state = task_cache_info.imperative_materialization_state
                object_lookup = state.object_lookup
                if id(x) in object_lookup:
                    # substitute imperatively materialized object references with
                    # their respective table objects
                    x = object_lookup[id(x)]
                if isinstance(x, (Table, RawSql)):
                    # fill assumed_dependencies for Tables that were not yet
                    # materialized
                    if len(state.assumed_dependencies) > 0:
                        if x.assumed_dependencies is None:
                            x.assumed_dependencies = list(
                                sorted(state.assumed_dependencies)
                            )
                return x

            result = deep_map(result, result_finalization_mutator)
            result = store.materialize_task(task, task_cache_info, result)

            # Delete underlying objects from result (after materializing them)
            def obj_del_mutator(x):
                if isinstance(x, (Table, Blob)):
                    x.obj = None
                if isinstance(x, RawSql):
                    x.loaded_tables = None
                if isinstance(x, (Table, RawSql)):
                    x.assumed_dependencies = deep_map(
                        x.assumed_dependencies, obj_del_mutator
                    )
                return x

            result = deep_map(result, obj_del_mutator)
            run_context.store_task_memo(task, memo_cache_key, result)
            self.value = result
            run_context.trace_hook.task_complete(task, result)

            return result

    def compute_auto_version(
        self,
        task: MaterializingTask,
        store: PipeDAGStore,
        bound: inspect.BoundArguments,
    ) -> str:
        # Decide on which type of auto versioning to use.
        # For this, we go off the `auto_version_support` flag of the retrieve
        # table hook of the input type
        from pydiverse.pipedag.backend.table.base import AutoVersionSupport

        hook = store.table_store.get_r_table_hook(task.input_type)
        auto_version_support = hook.auto_version_support

        if auto_version_support == AutoVersionSupport.NONE:
            raise TypeError(
                f"Auto versioning not supported for input type {task.input_type}."
            )

        # Perform auto versioning
        info = None
        if auto_version_support == AutoVersionSupport.LAZY:
            info = self.compute_auto_version_lazy(task, hook, store, bound)
        elif auto_version_support == AutoVersionSupport.TRACE:
            info = self.compute_auto_version_trace(task, hook, store, bound)

        assert isinstance(info, str)

        auto_version = "auto_" + stable_hash("AUTO-VERSION", info)
        return auto_version

    def compute_auto_version_lazy(
        self,
        task: MaterializingTask,
        hook: type[TableHook],
        store: PipeDAGStore,
        bound: inspect.BoundArguments,
    ) -> str:
        args, kwargs = store.dematerialize_task_inputs(
            task, bound.args, bound.kwargs, for_auto_versioning=True
        )

        try:
            result = self.fn(*args, **kwargs)
        except Exception as e:
            raise RuntimeError("Auto versioning failed") from e

        result, tables = self._auto_version_prepare_and_check_task_output(
            task, store, result
        )

        if len(tables) == 0:
            raise ValueError(
                f"Task with version={AUTO_VERSION} must return at least one Table."
            )

        # Compute auto version
        auto_version_info = []
        for table in tables:
            obj = table.obj
            auto_version_info.append(hook.get_auto_version_lazy(obj))

        return store.json_encode(
            {
                "task_output": result,
                "auto_version_info": auto_version_info,
            }
        )

    def compute_auto_version_trace(
        self,
        task: MaterializingTask,
        hook: type[TableHook],
        store: PipeDAGStore,
        bound: inspect.BoundArguments,
    ) -> str | None:
        computation_tracer = hook.get_computation_tracer()

        def tracer_proxy_mapper(x):
            if isinstance(x, Table):
                return computation_tracer.create_proxy(("TABLE", x.stage.name, x.name))
            if isinstance(x, RawSql):
                return computation_tracer.create_proxy(
                    ("RAW_SQL", x.stage.name, x.name, x.cache_key)
                )
            if isinstance(x, Blob):
                return computation_tracer.create_proxy(("BLOB", x.stage.name, x.name))

            return x

        args = deep_map(bound.args, tracer_proxy_mapper)
        kwargs = deep_map(bound.kwargs, tracer_proxy_mapper)

        with computation_tracer:
            try:
                result = self.fn(*args, **kwargs)
            except Exception as e:
                raise RuntimeError("Auto versioning failed") from e

        result, tables = self._auto_version_prepare_and_check_task_output(
            task, store, result
        )

        # Compute auto version
        trace_hash = computation_tracer.trace_hash()
        table_info = []

        # Include the computation tracers in the returned tables in the version
        # to invalidate the task when outputs get switched
        for table in tables:
            obj = table.obj
            assert isinstance(obj, ComputationTracerProxy)
            table_info.append(ComputationTraceRef(obj).id)

        return store.json_encode(
            {
                "task_output": result,
                "table_info": table_info,
                "trace_hash": trace_hash,
            }
        )

    def _auto_version_prepare_and_check_task_output(
        self, task: MaterializingTask, store: PipeDAGStore, output
    ):
        task_cache_info = TaskCacheInfo(
            task=task,
            input_hash="AUTO_VERSION",
            cache_fn_hash="AUTO_VERSION",
            cache_key="AUTO_VERSION",
            assert_no_materialization=False,
            force_task_execution=False,
        )

        # Replace computation tracer proxy objects in output
        def replace_proxy(x):
            if isinstance(x, ComputationTracerProxy):
                return {
                    "__pipedag_type__": "computation_tracer_proxy",
                    "id": ComputationTraceRef(x).id,
                }
            return x

        output = deep_map(output, replace_proxy)
        output, tables, raw_sqls, blobs = store.prepare_task_output_for_materialization(
            task, task_cache_info, output
        )

        if len(raw_sqls) != 0:
            raise ValueError(f"Task with version={AUTO_VERSION} can't return RawSql.")

        if len(blobs) != 0:
            raise ValueError(f"Task with version={AUTO_VERSION} can't return Blobs.")

        return output, tables


@overload
def input_stage_versions(
    *,
    name: str | None = None,
    input_type: type | None = None,
    version: str | None = None,
    cache: Callable[..., Any] | None = None,
    lazy: bool = False,
    group_node_tag: str | None = None,
    nout: int = 1,
    add_input_source: bool = False,
    ordering_barrier: bool | dict[str, Any] = True,
    call_context: Callable[[], Any] | None = None,
    include_views=True,
    lock_source_stages=True,
    pass_args: Iterable[str] = tuple(),
) -> Callable[[CallableT], CallableT | UnboundMaterializingTask]:
    ...


@overload
def input_stage_versions(fn: CallableT, /) -> CallableT | UnboundMaterializingTask:
    ...


def input_stage_versions(
    fn: CallableT | None = None,
    *,
    name: str | None = None,
    input_type: type | None = None,
    version: str | None = None,
    cache: Callable[..., Any] | None = None,
    lazy: bool = False,
    group_node_tag: str | None = None,
    nout: int = 1,
    add_input_source: bool = False,
    ordering_barrier: bool | dict[str, Any] = True,
    call_context: Callable[[], Any] | None = None,
    include_views=True,
    lock_source_stages=True,
    pass_args: Iterable[str] = tuple(),
):
    """
    A decorator that marks a function as a task that receives tables from two versions
    of the same stage. This can either be the currently active schema of this stage
    before schema swapping, or it can be an active schema of another pipeline instance.

    The arguments to the resulting task are significantly different from a task created
    with `@materialize` decorator. An arbitrary expression with lists, tuples, dicts,
    and Table/Blob/RawSQL references can be provided. This decorator will only collect
    tables referenced and will pass them as one dictionary per version to the task. It
    maps table names to some reference for the respective stage version in the form
    specified by input_type parameter.

    When a ConfigContext object is passed as toplevel parameter or named parameter in
    the wiring call of decorated task, it is used for referencing the second stage
    version. The idea is that the other configuration is just another pipeline instance
    accessible with the same SQLAlchemy engine of the table store. It may work also if
    this is not the case, but it may be harder for the task to process the dictionaries
    it receives. Currently, it is not allowed to pass more than one ConfigContext
    object.

    When no parameters (except for an optional ConfigContext) is passed, all currently
    existing tables of both stage versions (or their respective schema) are passed in
    the dictionaries to the task.

    With pass_blobs=True, two additional dictionaries are passed to the task at the end.
    In addition to the arguments mentioned above, pass_args can be used to feed defined
    keyword arguments through to the decorated task.

    :param fn:
        The function that gets executed by this task.
    :param name:
        The name of this task.
        If no `name` is provided, the name of `fn` is used instead.
    :param input_type:
        The data type as which to retrieve table objects from the store.
        All tables passed to this task get loaded from the table store and converted
        to this type. See :doc:`Table Backends </table_backends>` for
        more information.
    :param version:
        The version of this task.
        Unless the task is lazy, you always need to manually change this
        version number when you change the implementation to ensure that the
        task gets executed and the cache flushed.

        If the `version` is ``None`` and the task isn't lazy, then the task always
        gets executed, and all downstream tasks get invalidated.
    :param cache:
        An explicit cache function used to determine cache validity of the task inputs.

        This function gets called every time before the task gets executed.
        It gets called with the same arguments as the task.

        An explicit function for validating cache validity. If the output
        of this function changes while the source parameters are the same (e.g.
        the source is a filepath and `cache` loads data from this file), then the
        cache will be deemed invalid and is not used.
    :param lazy:
        Whether this task is lazy or not.

        Unlike a normal task, lazy tasks always get executed. However, if a lazy
        task produces a lazy table (e.g. a SQL query), the table store checks if
        the same query has been executed before. If this is the case, then the
        query doesn't get executed, and instead, the table gets copied from the cache.

        This behaviour is very useful, because you don't need to manually bump
        the `version` of a lazy task. This only works because for lazy tables
        generating the query is very cheap compared to executing it.
    :param group_node_tag:
        Set a tag that may add this task to a configuration based group node.
    :param nout:
        The number of objects returned by the task.
        If set, this allows unpacking and iterating over the results from the task.
    :param add_input_source:
        If true, Table and Blob objects are provided as tuple together with the
        dematerialized object. ``task(a=tbl)`` will result in
        ``a=(dematerialized_tbl, tbl)`` in the task.
    :param ordering_barrier:
        If true, the task will be surrounded by a GroupNode(ordering_barrier=True).
        If a dictionary is provided, it is surrounded by a
        GroupNode(**ordering_barrier). This allows passing style, style_tag, and label
        arguments.
        Attention: In contrast to @materialize, the default value is True.
    :param call_context:
        An optional context manager function that is opened before the task or its
        optional cache function is called and closed afterward.
    :param include_views:
        In case no explicit table references are given, if true, include views when
        collecting all tables of both stage versions.
    :param lock_source_stages:
        If true, lock the other stage version when a ConfigContext object is passed to
        this task.
    :param pass_args
        A list of named arguments that whould be passed from the call to the task
        function. By default, no arguments are passed, and just tables are extracted.

    Example
    -------

    ::

        @input_stage_versions(lazy=True, input_type=sa.Table)
        def validate_stage(
            transaction: dict[str, sa.Alias],
            other: dict[str, sa.Alias],
        ):
            compare_tables(transaction, other)

        def get_flow():
            with Flow() as f:
                a = produce_a_table()
                validate_stage()  # implicitly receives all tables written before
    """
    from pydiverse.pipedag import Stage

    forward_args = dict(
        name=name,
        input_type=input_type,
        version=version,
        cache=cache,
        lazy=lazy,
        group_node_tag=group_node_tag,
        nout=nout,
        add_input_source=add_input_source,
        ordering_barrier=ordering_barrier,
        call_context=call_context,
        include_views=include_views,
        lock_source_stages=lock_source_stages,
        pass_args=pass_args,
    )

    if fn is None:
        return partial(input_stage_versions, **forward_args)

    class OtherStageLockManager:
        def __init__(self):
            self.stages = None
            self.lock_manager = None

        @contextmanager
        def __call__(self):
            yield
            self.release_all()

        def release_all(self):
            # clear locks after either cache_fn or cache_fn and task function were
            # called
            if self.lock_manager is not None:
                for stage in sorted(self.stages, reverse=True):
                    self.lock_manager.release(stage.name)
                self.lock_manager.dispose()
                self.lock_manager = None

        def memorize(self, stages, lock_manager):
            self.release_all()
            self.stages = stages
            self.lock_manager = lock_manager

    other_stage_lock_manager = OtherStageLockManager()

    def cache_fn(*args, **kwargs):
        user_cache_fn = cache
        if user_cache_fn is None:
            # we don't care about the cache function, but we still need to cache the
            # arguments that are also used to call the task itself
            def user_cache_fn(*args, **kwargs):
                return None

        ret, cfg2, stages, lock_manager = _call_fn(
            user_cache_fn,
            args,
            kwargs,
            input_type=None,
            is_dematerialized=False,
            lock_source_stages=lock_source_stages,
        )
        other_stage_lock_manager.memorize(stages, lock_manager)
        if cfg2 is None:
            return uuid.uuid4().hex  # return random hash to ensure invalidation
        # get a hash of all task outputs of all relevant stage from other pipeline
        # instance
        ret = stable_hash(ret)
        for stage in stages:
            try:
                ret = ret + cfg2.store.table_store.get_stage_hash(stage)
            except sa.exc.ProgrammingError:
                # this happens for example if the other instance did not run, yet
                return uuid.uuid4().hex  # return random hash to ensure invalidation
        return ret

    # extract @materialize arguments from forward_args via __code__.co_varnames:
    n_args = materialize.__code__.co_argcount + materialize.__code__.co_kwonlyargcount
    arg_names = materialize.__code__.co_varnames[0:n_args]
    materialize_args = {k: v for k, v in forward_args.items() if k in arg_names}
    materialize_args["add_input_source"] = True  # we need the names
    materialize_args["cache"] = cache_fn
    if materialize_args.get("name") is None:
        materialize_args["name"] = getattr(fn, "__name__", "input_stage_versions-task")
    materialize_args["call_context"] = other_stage_lock_manager

    # For every input_state_versions-Task, we create a materialize-Task and give it the
    # same name as the input_state_versions-Task should get. The @materialize decorator
    # will dematerialize any Table/Blob/RawSQL objects passed to the
    # input_state_versions-Task. We pass add_input_source=True, so we still receive the
    # Table/Blob/RawSQL reference objects, so we can use them to dematerialize the
    # second stage version. Both are arranged in dictionaries indexable by table name or
    # blob name and handed over to the function annotated with @input_stage_versions.
    @materialize(**materialize_args)
    def task(*args, **kwargs):
        # the locks are already acquired by cache_fn if it is called
        lock_stages = (
            False
            if other_stage_lock_manager.lock_manager is not None
            else lock_source_stages
        )
        ret, _, _, _ = _call_fn(
            fn,
            args,
            kwargs,
            input_type=input_type,
            is_dematerialized=True,
            lock_source_stages=lock_stages,
        )
        return ret

    def _call_fn(
        _fn: Callable[..., Any],
        args: tuple,
        kwargs: dict[str, Any],
        *,
        input_type,
        is_dematerialized,
        lock_source_stages,
    ):
        _task = TaskContext.get().task
        stage = _task.stage
        pass_kwargs = {k: v for k, v in kwargs.items() if k in set(pass_args)}

        (
            stages,
            table_dict,
            blob_dict,
            raw_sqls,
            other_cfg,
            any_data_arguments_collected,
        ) = _collect_arguments(args, kwargs, is_dematerialized)
        stages.add(stage)

        # The connection to the other stage version configuration can either be the same
        # (in this case, we provide the active stage cache before swapping) or another
        # pipeline instance (in this case, we also take the active stage cache schema
        # of the same stage in the other instance)
        cfg1 = ConfigContext.get()
        cfg2 = other_cfg
        if cfg2 is None:
            cfg2 = cfg1
        stage2 = Stage(
            stage.name,
            stage.materialization_details,
            stage.group_node_tag,
            force_committed=True,
        )
        stage2.id = stage.id
        if lock_source_stages and other_cfg is not None:
            lock_manager = cfg2.create_lock_manager()
        else:
            lock_manager = None
            lock_source_stages = False

        transaction_table_dict, other_table_dict = {}, {}
        transaction_blob_dict, other_blob_dict = {}, {}
        try:
            # lock the source stages if they are read from another instance
            if lock_source_stages:
                for stage in sorted(stages):
                    lock_manager.acquire(stage.name)
            # cnt[0] counts the number of arguments in args and kwargs. pass_args within
            # kwargs are neglected.
            # cnt[0] is at least 2 due to deep_map calls on args and kwargs. If a
            # ConfigContext object is passed as argument, cfg_cnt=1 and cnt=3.
            if not any_data_arguments_collected:
                # dematerialize all tables and optionally views of both stage versions

                def _dematerialize_all(cfg, stage):
                    table_dict = {}
                    if include_views:
                        # also include table aliases
                        tbls1 = cfg.store.table_store.get_objects_in_stage(stage)
                    else:
                        # won't detect aliases
                        tbls1 = cfg.store.table_store.get_table_objects_in_stage(
                            stage, include_views=include_views
                        )
                    for name in tbls1:
                        tbl = Table(name=name)
                        tbl.stage = stage
                        if include_views:
                            (
                                tbl.name,
                                tbl.external_schema,
                            ) = cfg.store.table_store.resolve_alias(
                                tbl, stage.current_name
                            )
                            if not cfg.store.table_store.has_table_or_view(
                                tbl.name, tbl.external_schema
                            ):
                                continue  # skip because it is neither view nor alias
                        table_dict[tbl.name] = cfg.store.dematerialize_item(
                            tbl, input_type
                        )
                    return table_dict

                transaction_table_dict = _dematerialize_all(cfg1, stage)
                other_table_dict = _dematerialize_all(cfg2, stage2)
            else:

                def _dematerialize_versions(
                    ref: Table | Blob,
                    obj,
                    transaction_dict: dict[str, Any],
                    other_dict: dict[str, Any],
                    cfg2: ConfigContext,
                    stage: Stage,
                ):
                    key = (
                        ref.name
                        if ref.stage == stage
                        else f"{ref.stage.name}.{ref.name}"
                    )
                    transaction_dict[key] = obj
                    # load committed version of stage (either same pipeline instace or
                    # other)
                    tbl2 = ref.copy_without_obj()
                    tbl2.stage = Stage(
                        ref.stage.name,
                        ref.stage.materialization_details,
                        ref.stage.group_node_tag,
                        force_committed=True,
                    )
                    tbl2.stage.id = ref.stage.id
                    try:
                        store2 = cfg2.store.table_store
                        # Only try to dematerialize if an object with the correct name
                        # exists. Otherwise, we might run into retry loops.
                        if not isinstance(tbl2, Table) or tbl2.name.lower() in {
                            s.lower() for s in store2.get_objects_in_stage(tbl2.stage)
                        }:
                            other_dict[key] = cfg2.store.dematerialize_item(
                                tbl2, input_type
                            )
                        else:
                            other_dict[key] = (
                                "Table not found in other stage version: "
                                f"{PipedagJSONEncoder().encode(tbl2)}"
                            )
                    except Exception as e:
                        _task.logger.error(
                            "Failed to dematerialize table from other stage version",
                            table=ref,
                            stage=tbl2.stage,
                            instance=cfg2.instance_name,
                            instance_id=cfg2.instance_id,
                            cause=str(e),
                        )
                        other_dict[key] = e

                for tbl, obj in table_dict.items():
                    _dematerialize_versions(
                        tbl, obj, transaction_table_dict, other_table_dict, cfg2, stage
                    )
                for blob, obj in blob_dict.items():
                    _dematerialize_versions(
                        blob, obj, transaction_blob_dict, other_blob_dict, cfg2, stage
                    )
                for raw_sql in raw_sqls:
                    for tbl, obj in raw_sql.loaded_tables:
                        _dematerialize_versions(
                            tbl,
                            obj,
                            transaction_table_dict,
                            other_table_dict,
                            cfg2,
                            stage,
                        )

            optional_cfg = [] if other_cfg is None else [other_cfg]

            # restructure arbitrary arguments referencing Tables/Blobs/RawSQL into 2 or
            # 4 dictionaries providing references only to tables and optionally blobs
            if len(transaction_blob_dict) > 0:
                ret = _fn(
                    transaction_table_dict,
                    other_table_dict,
                    transaction_blob_dict,
                    other_blob_dict,
                    *optional_cfg,
                    **pass_kwargs,
                )
            else:
                ret = _fn(
                    transaction_table_dict,
                    other_table_dict,
                    *optional_cfg,
                    **pass_kwargs,
                )
        except Exception as e:
            if lock_source_stages:
                for stage in sorted(stages, reverse=True):
                    lock_manager.release(stage.name)
            raise e
        return ret, cfg2, stages, lock_manager

    def _collect_arguments(args, kwargs, is_dematerialized):
        remaining_kwargs = {k: v for k, v in kwargs.items() if k not in set(pass_args)}
        table_dict, blob_dict, raw_sqls = {}, {}, []
        stages = set()
        other_cfg = [None]  # list to allow modification in visitor
        cnt = [0]
        cfg_cnt = [0]
        str_dict_keys = [0]

        def visitor(x):
            if isinstance(x, dict):
                str_dict_keys[0] += sum(isinstance(k, str) for k in x.keys())
            if not isinstance(x, (list, tuple, dict)):
                cnt[0] += 1  # count non-collection arguments
            if isinstance(x, ConfigContext):
                if cfg_cnt[0] > 0:
                    raise ValueError(
                        "Only one ConfigContext object is allowed when calling a task "
                        "decorated with @input_stage_versions."
                    )
                other_cfg[0] = x
                cfg_cnt[0] += 1
            if is_dematerialized:
                if isinstance(x, tuple) and len(x) == 2:
                    # with the add_input_source parameter of @materialize, we receive
                    # tuples of the reference object (Table/Blob/RawSQL) together with
                    # the dematerialized object
                    obj, ref = x
                else:
                    return x
            else:
                obj, ref = x, x
            if isinstance(ref, (Table, Blob, RawSql)):
                stages.add(ref.stage)
            if isinstance(ref, Table):
                table_dict[ref] = obj
            elif isinstance(ref, Blob):
                blob_dict[ref] = obj
            elif isinstance(ref, RawSql):
                # ref and obj are identical for RawSql
                raw_sqls.append(ref)
            return x

        deep_map(args, visitor)
        deep_map(remaining_kwargs, visitor)
        # dictionary keys don't count since we like to be f(config=cfg) to be considered
        # as not any_data_arguements_collected
        any_data_arguements_collected = cnt[0] > cfg_cnt[0] + str_dict_keys[0]
        return (
            stages,
            table_dict,
            blob_dict,
            raw_sqls,
            other_cfg[0],
            any_data_arguements_collected,
        )

    return task


def _get_output_from_store(
    task: MaterializingTask | MaterializingTaskGetItem,
    as_type: type,
    ignore_position_hashes: bool = False,
) -> Any:
    """Helper to retrieve task output from store"""
    from pydiverse.pipedag.context.run_context import DematerializeRunContext
    from pydiverse.pipedag.materialize.store import dematerialize_output_from_store

    root_task = task if isinstance(task, Task) else task.task

    store = ConfigContext.get().store
    with DematerializeRunContext(root_task.flow):
        cached_output, _ = store.retrieve_most_recent_task_output_from_cache(
            root_task, ignore_position_hashes=ignore_position_hashes
        )
        return dematerialize_output_from_store(store, task, cached_output, as_type)


class PseudoStage:
    def __init__(self, stage: Stage, did_commit: bool):
        self._stage = stage
        self._name = stage.name
        self._transaction_name = stage.name
        self._did_commit = did_commit

        self.id = stage.id

    @property
    def name(self) -> str:
        return self._name

    @property
    def transaction_name(self) -> str:
        return self._transaction_name

    @property
    def current_name(self) -> str:
        if self._did_commit:
            return self.name
        else:
            return self.transaction_name


class AutoVersionType:
    """
    Special constant used to indicate that pipedag should automatically determine
    a version number for a task.

    The version is determined by running the task once to construct a representation
    of the computation performed on the tables (e.g. construct a query plan /
    computational graph).
    Using this representation, a unique version number is constructed such that
    if the computation changes the version number also changes.
    Then, if the task is deemed to be cache invalid, it is run again, but this
    time with actual data.

    This puts the following constraints on which tasks can use `AUTO_VERSION`:

    * The task must be a pure function.
    * The task may not inspect the contents of the input tables.
    * The task must return at least one table.
    * The task may not return :py:class:`RawSql`.
    * The task may not return :py:class:`Blob`.

    .. rubric:: Polars

    Automatic versioning is best supported by polars.
    To use it, you must specify
    :external+pl:doc:`polars.LazyFrame <reference/lazyframe/index>`
    as the task input type and only use LazyFrames inside your task.

    .. rubric:: Pandas

    Pandas support is still *experimental*. It is implemented using a technique
    we call "computation tracing", where we run the task with proxy tables
    as inputs that record all operations performed on them.
    This requires heavily monkey patching the pandas and numpy modules.
    As long as you only use pandas and numpy functions inside your task,
    computation tracing should work successfully.
    However, for the monkey patching to work, you only access pandas / numpy
    functions through their namespace (e.g. ``pd.concat(...)`` is allowed, while
    ``from pandas import concat; concat(...)`` is not allowed).

    Requires dask to be installed.


    Example
    -------
    ::

        @materialize(input_type=pl.LazyFrame, version=AUTO_VERSION)
        def polars_task(x: pl.LazyFrame):
            # Some operations that only utilize pl.LazyFrame...
            return x.with_columns(
                (pl.col("col1") * pl.col("col2")).alias("col3")
            )

    """

    def __str__(self):
        return "AUTO_VERSION"

    def __repr__(self):
        return "AUTO_VERSION"


AUTO_VERSION = AutoVersionType()



from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass
from enum import Enum
from typing import Any

import pandas as pd
import sqlalchemy as sa

from pydiverse.pipedag.backend.table.sql.ddl import (
    CreateTableWithSuffix,
    LockSourceTable,
    LockTable,
)
from pydiverse.pipedag.backend.table.sql.hooks import PandasTableHook
from pydiverse.pipedag.backend.table.sql.reflection import PipedagDB2Reflection
from pydiverse.pipedag.backend.table.sql.sql import SQLTableStore
from pydiverse.pipedag.backend.table.util import DType
from pydiverse.pipedag.container import Schema, Table
from pydiverse.pipedag.materialize.details import (
    BaseMaterializationDetails,
    resolve_materialization_details_label,
)

_TABLE_SPACE_KEYWORD_MAP = {
    "table_space_data": "IN",
    "table_space_index": "INDEX IN",
    "table_space_long": "LONG IN",
}


class IBMDB2CompressionTypes(str, Enum):
    NO_COMPRESSION = ""
    ROW_COMPRESSION = "COMPRESS YES"
    STATIC_ROW_COMPRESSION = "COMPRESS YES STATIC"
    ADAPTIVE_ROW_COMPRESSION = "COMPRESS YES ADAPTIVE"
    VALUE_COMPRESSION = "VALUE COMPRESSION"


@dataclass(frozen=True)
class IBMDB2MaterializationDetails(BaseMaterializationDetails):
    """
    :param compression: Specify the compression methods to be applied to the table.
        Possible values include
        compression="COMPRESS YES STATIC"
        compression="COMPRESS YES"
        compression="COMPRESS YES ADAPTIVE"
        compression="VALUE COMPRESSION"
        A list containing combining one of the first three with the last value, e.g.
        compression=["COMPRESS YES ADAPTIVE", "VALUE COMPRESSION"]

        compression="" will result in no compression
    :param table_space_data: The DB2 table space where the data is  stored.
    :param table_space_index: The DB2 table space where the partitioned index is stored.
    :param table_space_long: The DB2 table spaces where the values of any long columns
        are stored.
    """

    def __post_init__(self):
        object.__setattr__(
            self,
            "compression",
            IBMDB2CompressionTypes(self.compression)
            if isinstance(self.compression, str)
            else [IBMDB2CompressionTypes(c) for c in self.compression]
            if self.compression is not None
            else None,
        )

    compression: IBMDB2CompressionTypes | list[IBMDB2CompressionTypes] | None = None
    table_space_data: str | None = None
    table_space_index: str | None = None
    table_space_long: str | list[str] | None = None


class IBMDB2TableStore(SQLTableStore):
    """
    SQLTableStore that supports `IBM Db2 <https://www.ibm.com/products/db2>`_.
    Requires `ibm-db-sa <https://pypi.org/project/ibm-db-sa/>`_ to be installed.

    Takes the same arguments as
    :py:class:`SQLTableStore <pydiverse.pipedag.backend.table.SQLTableStore>`
    """

    _dialect_name = "ibm_db_sa"

    def _default_isolation_level(self):
        return "UNCOMMITTED READ"

    def lock_table(
        self, table: Table | str, schema: Schema | str, conn: Any = None
    ) -> list:
        """
        For some dialects, it might be beneficial to lock a table before writing to it.
        """
        stmt = LockTable(table.name if isinstance(table, Table) else table, schema)
        if conn is not None:
            self.execute(stmt, conn=conn)
        return [stmt]

    def lock_source_table(
        self, table: Table | str, schema: Schema | str, conn: Any = None
    ) -> list:
        """
        For some dialects, it might be beneficial to lock source tables before reading.
        """
        stmt = LockSourceTable(
            table.name if isinstance(table, Table) else table, schema
        )
        if conn is not None:
            self.execute(stmt, conn=conn)
        return [stmt]

    def dialect_requests_empty_creation(self, table: Table, is_sql: bool) -> bool:
        if is_sql:
            # IBM DB2 does not support CREATE TABLE AS SELECT without INSERT INTO
            return True
        else:
            label = resolve_materialization_details_label(table)
            return (
                (label is not None and len(label.strip()) > 0)
                or table.nullable is not None
                or table.non_nullable is not None
                or (table.primary_key is not None and len(table.primary_key) > 0)
            )

    def get_forced_nullability_columns(
        self, table: Table, table_cols: Iterable[str], report_nullable_cols=False
    ) -> tuple[list[str], list[str]]:
        # ibm_db2 dialect has literals as non-nullable types by default, so we also need
        # the list of nullable columns to fix
        nullable_cols, non_nullable_cols = self._process_table_nullable_parameters(
            table, table_cols
        )
        # add primery key columns to non_nullable_cols
        if table.primary_key:
            primary_key = (
                table.primary_key
                if isinstance(table.primary_key, list)
                else [table.primary_key]
            )
            non_nullable_cols += primary_key
            nullable_cols = [
                col for col in nullable_cols if col not in table.primary_key
            ]
        return nullable_cols, non_nullable_cols

    def add_indexes_and_set_nullable(
        self,
        table: Table,
        schema: Schema,
        *,
        on_empty_table: bool | None = None,
        table_cols: Iterable[str] | None = None,
    ):
        super().add_indexes_and_set_nullable(
            table, schema, on_empty_table=on_empty_table, table_cols=table_cols
        )
        table_name = self.engine.dialect.identifier_preparer.quote(table.name)
        schema_name = self.engine.dialect.identifier_preparer.quote_schema(schema.get())
        query = (
            f"CALL SYSPROC.ADMIN_CMD('RUNSTATS ON TABLE {schema_name}.{table_name}"
            f" ON ALL COLUMNS WITH DISTRIBUTION ON ALL COLUMNS AND UNSAMPLED"
            f" DETAILED INDEXES ALL SET PROFILE');"
        )
        self.execute(query)

    def resolve_alias(self, table: Table, stage_name: str) -> tuple[str, str]:
        # The base implementation already takes care of converting Table objects
        # based on ExternalTableReference objects to string table name and schema.
        # For normal Table objects, it needs the stage schema name.
        table_name, schema = super().resolve_alias(table, stage_name)
        return PipedagDB2Reflection.resolve_alias(self.engine, table_name, schema)

    def check_materialization_details_supported(self, label: str | None) -> None:
        _ = label
        return

    def _set_materialization_details(
        self, materialization_details: dict[str, dict[str | list[str]]] | None
    ) -> None:
        self.materialization_details = (
            IBMDB2MaterializationDetails.create_materialization_details_dict(
                materialization_details,
                self.strict_materialization_details,
                self.default_materialization_details,
                self.logger,
            )
        )

    def _get_compression(
        self, materialization_details_label: str | None
    ) -> str | list[str] | None:
        compression: IBMDB2CompressionTypes | list[
            IBMDB2CompressionTypes
        ] | None = IBMDB2MaterializationDetails.get_attribute_from_dict(
            self.materialization_details,
            materialization_details_label,
            self.default_materialization_details,
            "compression",
            self.strict_materialization_details,
            self.logger,
        )
        if isinstance(compression, list):
            return [c.value for c in compression]
        if compression is not None:
            return compression.value

    def _get_table_spaces(
        self, materialization_details_label: str | None
    ) -> dict[str, str]:
        return {
            f"table_space_{st}": IBMDB2MaterializationDetails.get_attribute_from_dict(
                self.materialization_details,
                materialization_details_label,
                self.default_materialization_details,
                f"table_space_{st}",
                self.strict_materialization_details,
                self.logger,
            )
            for st in ("data", "index", "long")
        }

    def get_create_table_suffix(self, materialization_details_label: str | None) -> str:
        table_spaces = self._get_table_spaces(materialization_details_label)
        table_space_suffix = " ".join(
            f"{_TABLE_SPACE_KEYWORD_MAP[stype]} "
            f"{self.engine.dialect.identifier_preparer.quote(sname)}"
            for stype, sname in table_spaces.items()
            if sname
        )
        compression = self._get_compression(materialization_details_label)
        if isinstance(compression, str):
            compression = [compression]
        elif compression is None:
            compression = []
        compression_suffix = " ".join(compression)
        return " ".join((table_space_suffix, compression_suffix))

    def _get_all_objects_in_schema(self, schema: Schema) -> dict[str, Any]:
        return PipedagDB2Reflection.get_all_objects(self.engine, schema.get())


@IBMDB2TableStore.register_table(pd)
class PandasTableHook(PandasTableHook):
    @classmethod
    def _get_dialect_dtypes(cls, dtypes: dict[str, DType], table: Table[pd.DataFrame]):
        # Default string target is CLOB which can't be used for indexing.
        # -> Convert indexed string columns to VARCHAR(256)
        index_columns = set()
        if indexes := table.indexes:
            index_columns |= {col for index in indexes for col in index}
        if primary_key := table.primary_key:
            index_columns |= set(primary_key)

        return ({name: dtype.to_sql() for name, dtype in dtypes.items()}) | (
            {
                name: (
                    sa.String(length=256)
                    if name in index_columns
                    else sa.String(length=32_672)
                )
                for name, dtype in dtypes.items()
                if dtype == DType.STRING
            }
        )

    @classmethod
    def _dialect_create_empty_table(
        cls,
        store: SQLTableStore,
        df: pd.DataFrame,
        table: Table[pd.DataFrame],
        schema: Schema,
        dtypes: dict[str, DType],
    ):
        suffix = store.get_create_table_suffix(
            resolve_materialization_details_label(table)
        )
        store.execute(
            CreateTableWithSuffix(
                table.name,
                schema,
                dtypes,
                table.nullable,
                table.non_nullable,
                suffix,
            )
        )



from __future__ import annotations

import sqlalchemy as sa

from pydiverse.pipedag.backend.table.util import DType


def test_dtype_from_sqlalchemy():
    def assert_conversion(type_, expected):
        assert DType.from_sql(type_) == expected

    assert_conversion(sa.BigInteger(), DType.INT64)
    assert_conversion(sa.Integer(), DType.INT32)
    assert_conversion(sa.SmallInteger(), DType.INT16)

    assert_conversion(sa.Numeric(), DType.FLOAT64)
    assert_conversion(sa.Numeric(13, 2), DType.FLOAT64)
    assert_conversion(sa.Numeric(1, 0), DType.FLOAT64)
    assert_conversion(sa.DECIMAL(13, 2), DType.FLOAT64)
    assert_conversion(sa.DECIMAL(1, 0), DType.FLOAT64)
    assert_conversion(sa.Float(), DType.FLOAT64)
    assert_conversion(sa.Float(24), DType.FLOAT32)
    assert_conversion(sa.Float(53), DType.FLOAT64)

    assert_conversion(sa.String(), DType.STRING)
    assert_conversion(sa.Boolean(), DType.BOOLEAN)

    assert_conversion(sa.Date(), DType.DATE)
    assert_conversion(sa.Time(), DType.TIME)
    assert_conversion(sa.DateTime(), DType.DATETIME)


def test_dtype_to_sqlalchemy():
    def assert_conversion(type_: DType, expected):
        assert isinstance(type_.to_sql(), expected)

    assert_conversion(DType.INT64, sa.BigInteger)
    assert_conversion(DType.INT32, sa.Integer)
    assert_conversion(DType.INT16, sa.SmallInteger)
    assert_conversion(DType.INT8, sa.SmallInteger)

    assert_conversion(DType.UINT64, sa.BigInteger)
    assert_conversion(DType.UINT32, sa.BigInteger)
    assert_conversion(DType.UINT16, sa.Integer)
    assert_conversion(DType.UINT8, sa.SmallInteger)

    assert_conversion(DType.FLOAT64, sa.Float)
    assert_conversion(DType.FLOAT32, sa.Float)

    assert_conversion(DType.STRING, sa.String)
    assert_conversion(DType.BOOLEAN, sa.Boolean)

    assert_conversion(DType.DATE, sa.Date)
    assert_conversion(DType.TIME, sa.Time)
    assert_conversion(DType.DATETIME, sa.DateTime)



from __future__ import annotations

import datetime as dt
import string
from typing import Any

import pandas as pd
import pyarrow as pa
import pytest
import sqlalchemy as sa
from packaging.version import Version

import tests.util.tasks_library as m
from pydiverse.pipedag import *
from pydiverse.pipedag.backend.table.sql.hooks import PandasTableHook
from pydiverse.pipedag.backend.table.sql.sql import DISABLE_DIALECT_REGISTRATION
from pydiverse.pipedag.backend.table.util import DType

# Parameterize all tests in this file with several instance_id configurations
from tests.fixtures.instances import DATABASE_INSTANCES, with_instances
from tests.util.spy import spy_task
from tests.util.sql import get_config_with_table_store

# disable duckdb for now, since they have a bug in version 0.9.2 that needs fixing
pytestmark = [with_instances(tuple(set(DATABASE_INSTANCES) - {"duckdb"}))]
pd_version = Version(pd.__version__)


class Values:
    INT8 = [-128, 127]
    INT16 = [-32768, 32767]
    INT32 = [-2147483648, 2147483647]
    INT64 = [-9223372036854775808, 9223372036854775807]

    FLOAT32 = [3e38, 3.14159]
    FLOAT64 = [1e300, 3.14159]

    STR = ["", string.printable + string.whitespace]

    BOOLEAN = [True, False]

    DATE = [dt.date(1, 1, 1), dt.date(9999, 12, 31)]
    DATE_NS = [dt.date(1700, 1, 1), dt.date(2200, 1, 1)]
    TIME = [dt.time(0, 0, 0), dt.time(23, 59, 59)]
    DATETIME = [dt.datetime(1, 1, 1, 0, 0, 0), dt.datetime(9999, 12, 31, 23, 59, 59)]
    DATETIME_NS = [dt.datetime(1700, 1, 1), dt.datetime(2200, 1, 1)]


class NoneValues:
    INT = [0, None]
    FLOAT = [0.0, None]
    STR = ["", None]
    BOOLEAN = [False, None]

    DATE = [dt.date(1970, 1, 1), None]
    TIME = [dt.time(0, 0, 0), None]
    DATETIME = [dt.datetime(1970, 1, 1, 0, 0, 0), None]


def get_dialect_name():
    return ConfigContext.get().store.table_store.engine.dialect.name


class TestPandasTableHookNumpy:
    def test_basic(self):
        df = pd.DataFrame(
            {
                "int8": pd.array(Values.INT8, dtype="Int8"),
                "int16": pd.array(Values.INT16, dtype="Int16"),
                "int32": pd.array(Values.INT32, dtype="Int32"),
                "int64": pd.array(Values.INT64, dtype="Int64"),
                "float32": pd.array(Values.FLOAT32, dtype="Float32"),
                "float64": pd.array(Values.FLOAT32, dtype="Float64"),
                "str": pd.array(Values.STR, dtype="string"),
                "boolean": pd.array(Values.BOOLEAN, dtype="boolean"),
            }
        )

        @materialize()
        def numpy_input():
            return df

        @materialize(input_type=(pd.DataFrame, "numpy"))
        def assert_expected(in_df):
            pd.testing.assert_frame_equal(in_df, df, check_dtype=False)

            allowed_dtypes = set(df.dtypes)
            for dtype in in_df.dtypes:
                assert dtype in allowed_dtypes

        with Flow() as f:
            with Stage("stage"):
                t = numpy_input()
                assert_expected(t)

        assert f.run().successful

    def test_none(self):
        df = pd.DataFrame(
            {
                "int": pd.array(NoneValues.INT, dtype="Int64"),
                "float": pd.array(NoneValues.FLOAT, dtype="Float64"),
                "str": pd.array(NoneValues.STR, dtype="string"),
                "boolean": pd.array(NoneValues.BOOLEAN, dtype="boolean"),
            }
        )

        @materialize()
        def numpy_input():
            return df

        @materialize(input_type=(pd.DataFrame, "numpy"))
        def assert_expected(in_df):
            pd.testing.assert_frame_equal(in_df, df, check_dtype=False)

            allowed_dtypes = set(df.dtypes)
            for col, dtype in in_df.dtypes.items():
                if col == "boolean" and get_dialect_name() == "ibm_db_sa":
                    assert dtype == pd.Int16Dtype()
                else:
                    assert dtype in allowed_dtypes

        with Flow() as f:
            with Stage("stage"):
                t = numpy_input()
                assert_expected(t)

        assert f.run().successful

    def test_datetime(self):
        df = pd.DataFrame(
            {
                "date": Values.DATE,
                "date_ns": Values.DATE_NS,
                "date_none": NoneValues.DATE,
                "datetime": Values.DATETIME,
                "datetime_ns": Values.DATETIME_NS,
                "datetime_none": NoneValues.DATETIME,
            }
        ).astype(object)

        df_expected = pd.DataFrame(
            {
                "date": pd.array(Values.DATE_NS, dtype="datetime64[ns]"),
                "date_ns": pd.array(Values.DATE_NS, dtype="datetime64[ns]"),
                "date_none": pd.array(NoneValues.DATE, dtype="datetime64[ns]"),
                "datetime": pd.array(Values.DATETIME_NS, dtype="datetime64[ns]"),
                "datetime_ns": pd.array(Values.DATETIME_NS, dtype="datetime64[ns]"),
                "datetime_none": pd.array(NoneValues.DATETIME, dtype="datetime64[ns]"),
                "date_year": [d.year for d in Values.DATE],
                "date_ns_year": [d.year for d in Values.DATE_NS],
                "date_none_year": [1970, None],
                "datetime_year": [d.year for d in Values.DATETIME],
                "datetime_ns_year": [d.year for d in Values.DATETIME_NS],
                "datetime_none_year": [1970, None],
            }
        )

        @materialize()
        def numpy_input():
            datetime_dtype = (
                sa.DateTime()
                if (get_dialect_name() != "mssql")
                else sa.dialects.mssql.DATETIME2()
            )

            return Table(
                df,
                type_map={
                    "date": sa.Date(),
                    "date_ns": sa.Date(),
                    "date_none": sa.Date(),
                    "datetime": datetime_dtype,
                    "datetime_ns": datetime_dtype,
                    "datetime_none": datetime_dtype,
                },
            )

        @materialize(input_type=(pd.DataFrame, "numpy"))
        def assert_expected(in_df):
            pd.testing.assert_frame_equal(in_df, df_expected, check_dtype=False)

        with Flow() as f:
            with Stage("stage"):
                t = numpy_input()
                assert_expected(t)

        assert f.run().successful


@pytest.mark.skipif(pd_version < Version("2.0"), reason="Requires pandas v2.0")
class TestPandasTableHookArrow:
    def test_basic(self):
        df = pd.DataFrame(
            {
                "int8": pd.array(Values.INT8, dtype="int8[pyarrow]"),
                "int16": pd.array(Values.INT16, dtype="int16[pyarrow]"),
                "int32": pd.array(Values.INT32, dtype="int32[pyarrow]"),
                "int64": pd.array(Values.INT64, dtype="int64[pyarrow]"),
                "float32": pd.array(Values.FLOAT32, dtype="float[pyarrow]"),
                "float64": pd.array(Values.FLOAT32, dtype="double[pyarrow]"),
                "str": pd.array(Values.STR, dtype=pd.ArrowDtype(pa.string())),
                "boolean": pd.array(Values.BOOLEAN, dtype="bool[pyarrow]"),
                "date": pd.array(Values.DATE, dtype=pd.ArrowDtype(pa.date32())),
                "time": pd.array(Values.TIME, dtype=pd.ArrowDtype(pa.time64("us"))),
                "datetime": pd.array(
                    Values.DATETIME, dtype=pd.ArrowDtype(pa.timestamp("us"))
                ),
            }
        )

        @materialize()
        def arrow_input():
            return df

        @materialize(input_type=(pd.DataFrame, "arrow"))
        def assert_expected(in_df):
            pd.testing.assert_frame_equal(in_df, df, check_dtype=False)
            allowed_dtypes = set(df.dtypes)
            # Prefer StringDtype("pyarrow") over ArrowDtype(pa.string()) for now.
            # We need to check this choice with future versions of pandas/pyarrow.
            allowed_dtypes.remove(pd.ArrowDtype(pa.string()))
            allowed_dtypes.add(pd.StringDtype("pyarrow"))
            for col, dtype in in_df.dtypes.items():
                if col == "boolean" and get_dialect_name() == "ibm_db_sa":
                    assert dtype == pd.ArrowDtype(pa.int16())
                else:
                    assert dtype in allowed_dtypes

        with Flow() as f:
            with Stage("stage"):
                t = arrow_input()
                assert_expected(t)

        assert f.run().successful

    def test_none(self):
        df = pd.DataFrame(
            {
                "int": pd.array(NoneValues.INT, dtype="int64[pyarrow]"),
                "float": pd.array(NoneValues.FLOAT, dtype="double[pyarrow]"),
                "str": pd.array(NoneValues.STR, dtype=pd.ArrowDtype(pa.string())),
                "boolean": pd.array(NoneValues.BOOLEAN, dtype="bool[pyarrow]"),
                "date": pd.array(NoneValues.DATE, dtype=pd.ArrowDtype(pa.date32())),
                "time": pd.array(NoneValues.TIME, dtype=pd.ArrowDtype(pa.time64("us"))),
                "datetime": pd.array(
                    NoneValues.DATETIME, dtype=pd.ArrowDtype(pa.timestamp("us"))
                ),
            }
        )

        @materialize()
        def arrow_input():
            return df

        @materialize(input_type=(pd.DataFrame, "arrow"))
        def assert_expected(in_df):
            pd.testing.assert_frame_equal(in_df, df, check_dtype=False)

        with Flow() as f:
            with Stage("stage"):
                t = arrow_input()
                assert_expected(t)

        assert f.run().successful


@pytest.mark.xfail(
    reason=(
        "The string '\\N' can't be materialized using the "
        "COPY FROM STDIN WITH CSV technique."
    ),
    strict=True,
)
@with_instances("postgres")
def test_pandas_table_hook_postgres_null_string():
    data = {"strNA": ["", None, "\\N"]}

    with Flow() as f:
        with Stage("stage_0"):
            t = m.pd_dataframe(data)

    with ConfigContext.get().evolve(swallow_exceptions=True), StageLockContext():
        result = f.run()
        df = result.get(t, as_type=pd.DataFrame)

    assert df["strNA"][0] == ""
    assert pd.isna(df["strNA"][1])
    assert (
        df["strNA"].fillna("X")[2] == "\\N"
    ), "This is a known issue that the string '\\N' is considered NA"


@with_instances("postgres", "local_table_store")
class TestPandasAutoVersion:
    def test_smoke(self, mocker):
        should_swap_inputs = False
        global_df = pd.DataFrame({"global_col": [1, 0, 1, 0]})

        @materialize(input_type=pd.DataFrame, version=AUTO_VERSION, nout=2)
        def in_tables():
            in_table_1 = pd.DataFrame({"col": [1, 2, 3, 4]})
            in_table_2 = pd.DataFrame({"col": [4, 3, 2, 1]})

            if should_swap_inputs:
                return Table(in_table_2), Table(in_table_1)
            else:
                return Table(in_table_1), Table(in_table_2)

        @materialize(input_type=pd.DataFrame, version=AUTO_VERSION)
        def noop(df):
            return df, Table(df)

        @materialize(input_type=pd.DataFrame, version=AUTO_VERSION)
        def global_df_task(df):
            df["col"] *= global_df["global_col"]
            return df

        with Flow() as f:
            with Stage("stage"):
                in_tables_ = in_tables()
                noop_ = noop(in_tables_[0])
                global_df_task(noop_[0])

        f.run()
        in_tables_spy = spy_task(mocker, in_tables_)

        f.run()
        in_tables_spy.assert_called_once()

        should_swap_inputs = True
        f.run()
        in_tables_spy.assert_called(2)


@with_instances("postgres")
class TestPandasCustomHook:
    def test_custom_upload(self):
        class TestTableStore(ConfigContext.get().store.table_store.__class__):
            _dialect_name = DISABLE_DIALECT_REGISTRATION
            # this subclass is just to make sure hooks of other tests are not affected
            pass

        class TestTableStore2(TestTableStore):
            # this tests that the delegation to parent hooks works
            pass

        cfg = get_config_with_table_store(ConfigContext.get(), TestTableStore2)

        @TestTableStore.register_table(pd, replace_hooks=[PandasTableHook])
        class CustomPandasUploadTableHook(PandasTableHook):
            @classmethod
            def upload_table(
                cls,
                df: pd.DataFrame,
                name: str,
                schema: str,
                dtypes: dict[str, DType],
                conn: sa.Connection,
                early: bool,
            ):
                df["custom_upload"] = True
                super().upload_table(df, name, schema, dtypes, conn, early)

        df = pd.DataFrame(
            {
                "int8": pd.array(Values.INT8, dtype="Int8"),
            }
        )

        @materialize()
        def numpy_input():
            return df

        @materialize(input_type=pd.DataFrame)
        def verify_cutom(t):
            assert "custom_upload" in t.columns

        with Flow() as f:
            with Stage("stage"):
                t = numpy_input()
                verify_cutom(t)

        assert f.run(config=cfg).successful

    def test_custom_download(self):
        class TestTableStore(ConfigContext.get().store.table_store.__class__):
            _dialect_name = DISABLE_DIALECT_REGISTRATION
            # this subclass is just to make sure hooks of other tests are not affected
            pass

        cfg = get_config_with_table_store(ConfigContext.get(), TestTableStore)

        @TestTableStore.register_table(pd, replace_hooks=[PandasTableHook])
        class CustomPandasDownloadTableHook(PandasTableHook):
            @classmethod
            def download_table(
                cls,
                query: Any,
                conn: sa.Connection,
                dtypes: dict[str, DType] | None = None,
            ):
                df = super().download_table(query, conn, dtypes)
                df["custom_download"] = True
                return df

        df = pd.DataFrame(
            {
                "int8": pd.array(Values.INT8, dtype="Int8"),
            }
        )

        @materialize()
        def numpy_input():
            return df

        @materialize(input_type=pd.DataFrame)
        def verify_cutom(t):
            assert "custom_download" in t.columns

        with Flow() as f:
            with Stage("stage"):
                t = numpy_input()
                verify_cutom(t)

        assert f.run(config=cfg).successful



from __future__ import annotations

import datetime as dt

import numpy as np
import pandas as pd
import pyarrow as pa
import pytest

from pydiverse.pipedag.backend.table.util import DType, PandasDTypeBackend


def test_dtype_from_pandas():
    def assert_conversion(type_, expected):
        assert DType.from_pandas(type_) == expected

    assert_conversion(int, DType.INT64)
    assert_conversion(float, DType.FLOAT64)
    assert_conversion(str, DType.STRING)
    assert_conversion(bool, DType.BOOLEAN)

    # Objects should get converted to string
    assert_conversion(object, DType.STRING)
    assert_conversion(dt.date, DType.STRING)
    assert_conversion(dt.time, DType.STRING)
    assert_conversion(dt.datetime, DType.STRING)

    # Numpy types
    assert_conversion(np.int64, DType.INT64)
    assert_conversion(np.int32, DType.INT32)
    assert_conversion(np.int16, DType.INT16)
    assert_conversion(np.int8, DType.INT8)

    assert_conversion(np.uint64, DType.UINT64)
    assert_conversion(np.uint32, DType.UINT32)
    assert_conversion(np.uint16, DType.UINT16)
    assert_conversion(np.uint8, DType.UINT8)

    assert_conversion(np.floating, DType.FLOAT64)
    assert_conversion(np.float64, DType.FLOAT64)
    assert_conversion(np.float32, DType.FLOAT32)

    assert_conversion(np.bytes_, DType.STRING)
    assert_conversion(np.bool_, DType.BOOLEAN)

    assert_conversion(np.datetime64, DType.DATETIME)
    assert_conversion(np.dtype("datetime64[ms]"), DType.DATETIME)
    assert_conversion(np.dtype("datetime64[ns]"), DType.DATETIME)

    # Numpy nullable extension types
    assert_conversion(pd.Int64Dtype(), DType.INT64)
    assert_conversion(pd.Int32Dtype(), DType.INT32)
    assert_conversion(pd.Int16Dtype(), DType.INT16)
    assert_conversion(pd.Int8Dtype(), DType.INT8)

    assert_conversion(pd.UInt64Dtype(), DType.UINT64)
    assert_conversion(pd.UInt32Dtype(), DType.UINT32)
    assert_conversion(pd.UInt16Dtype(), DType.UINT16)
    assert_conversion(pd.UInt8Dtype(), DType.UINT8)

    assert_conversion(pd.Float64Dtype(), DType.FLOAT64)
    assert_conversion(pd.Float32Dtype(), DType.FLOAT32)

    assert_conversion(pd.StringDtype(), DType.STRING)
    assert_conversion(pd.BooleanDtype(), DType.BOOLEAN)


def test_dtype_to_pandas_numpy():
    def assert_conversion(type_: DType, expected):
        assert type_.to_pandas(PandasDTypeBackend.NUMPY) == expected

    assert_conversion(DType.INT64, pd.Int64Dtype())
    assert_conversion(DType.INT32, pd.Int32Dtype())
    assert_conversion(DType.INT16, pd.Int16Dtype())
    assert_conversion(DType.INT8, pd.Int8Dtype())

    assert_conversion(DType.UINT64, pd.UInt64Dtype())
    assert_conversion(DType.UINT32, pd.UInt32Dtype())
    assert_conversion(DType.UINT16, pd.UInt16Dtype())
    assert_conversion(DType.UINT8, pd.UInt8Dtype())

    assert_conversion(DType.STRING, pd.StringDtype())
    assert_conversion(DType.BOOLEAN, pd.BooleanDtype())

    assert_conversion(DType.DATE, np.dtype("datetime64[ns]"))
    assert_conversion(DType.DATETIME, np.dtype("datetime64[ns]"))

    with pytest.raises(TypeError):
        DType.TIME.to_pandas(PandasDTypeBackend.NUMPY)


@pytest.mark.skipif("pd.__version__ < '2'")
def test_dtype_to_pandas_pyarrow():
    def assert_conversion(type_: DType, expected):
        if isinstance(expected, pa.DataType):
            assert type_.to_pandas(PandasDTypeBackend.ARROW) == pd.ArrowDtype(expected)
        else:
            assert type_.to_pandas(PandasDTypeBackend.ARROW) == expected

    assert_conversion(DType.INT64, pa.int64())
    assert_conversion(DType.INT32, pa.int32())
    assert_conversion(DType.INT16, pa.int16())
    assert_conversion(DType.INT8, pa.int8())

    assert_conversion(DType.UINT64, pa.uint64())
    assert_conversion(DType.UINT32, pa.uint32())
    assert_conversion(DType.UINT16, pa.uint16())
    assert_conversion(DType.UINT8, pa.uint8())

    assert_conversion(DType.STRING, pd.StringDtype(storage="pyarrow"))
    assert_conversion(DType.BOOLEAN, pa.bool_())

    assert_conversion(DType.DATE, pa.date32())
    assert_conversion(DType.TIME, pa.time64("us"))
    assert_conversion(DType.DATETIME, pa.timestamp("us"))



from __future__ import annotations

import tempfile

import pandas as pd
import sqlalchemy as sa

from pydiverse.pipedag import Flow, Stage, Table, materialize
from pydiverse.pipedag.context import StageLockContext
from pydiverse.pipedag.core.config import create_basic_pipedag_config
from pydiverse.pipedag.util.structlog import setup_logging


@materialize(lazy=True)
def lazy_task_1():
    return sa.select(
        sa.literal(1).label("x"),
        sa.literal(2).label("y"),
    )


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_2(input1: sa.sql.expression.Alias, input2: sa.sql.expression.Alias):
    query = sa.select(
        (input1.c.x * 5).label("x5"),
        input2.c.a,
    ).select_from(input1.outerjoin(input2, input2.c.x == input1.c.x))

    return Table(query, name="task_2_out", primary_key=["a"])


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_3(input1: sa.sql.expression.Alias):
    return sa.text(f"SELECT * FROM {input1.original.schema}.{input1.original.name}")


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_4(input1: sa.sql.expression.Alias):
    return sa.text(f"SELECT * FROM {input1.original.schema}.{input1.original.name}")


@materialize(nout=2, version="1.0.0")
def eager_inputs():
    dfA = pd.DataFrame(
        {
            "a": [0, 1, 2, 4],
            "b": [9, 8, 7, 6],
        }
    )
    dfB = pd.DataFrame(
        {
            "a": [2, 1, 0, 1],
            "x": [1, 1, 2, 2],
        }
    )
    return Table(dfA, "dfA"), Table(dfB, "dfB_%%")


@materialize(version="1.0.0", input_type=pd.DataFrame)
def eager_task(tbl1: pd.DataFrame, tbl2: pd.DataFrame):
    return tbl1.merge(tbl2, on="x")


def main():
    with tempfile.TemporaryDirectory() as temp_dir:
        cfg = create_basic_pipedag_config(
            f"duckdb:///{temp_dir}/db.duckdb",
            disable_stage_locking=True,  # This is special for duckdb
            # Attention: If uncommented, stage and task names might be sent to the
            #   following URL. You can self-host kroki if you like:
            #   https://docs.kroki.io/kroki/setup/install/
            # kroki_url="https://kroki.io",
        ).get("default")
        with cfg:
            with Flow() as f:
                with Stage("stage_1"):
                    lazy_1 = lazy_task_1()
                    a, b = eager_inputs()

                with Stage("stage_2"):
                    lazy_2 = lazy_task_2(lazy_1, b)
                    lazy_3 = lazy_task_3(lazy_2)
                    eager = eager_task(lazy_1, b)

                with Stage("stage_3"):
                    lazy_4 = lazy_task_4(lazy_2)
                _ = lazy_3, lazy_4, eager  # unused terminal output tables

            # Run flow
            result = f.run()
            assert result.successful

            # Run in a different way for testing
            with StageLockContext():
                result = f.run()
                assert result.successful
                assert result.get(lazy_1, as_type=pd.DataFrame)["x"][0] == 1


if __name__ == "__main__":
    setup_logging()  # you can setup the logging and/or structlog libraries as you wish
    main()



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from pydiverse.pipedag.backend.table.util import DType

pl = pytest.importorskip("polars")

if TYPE_CHECKING:
    import polars as pl


def test_dtype_from_polars():
    def assert_conversion(type_, expected):
        assert DType.from_polars(type_) == expected

    assert_conversion(pl.Int64, DType.INT64)
    assert_conversion(pl.Int32, DType.INT32)
    assert_conversion(pl.Int16, DType.INT16)
    assert_conversion(pl.Int8, DType.INT8)

    assert_conversion(pl.UInt64, DType.UINT64)
    assert_conversion(pl.UInt32, DType.UINT32)
    assert_conversion(pl.UInt16, DType.UINT16)
    assert_conversion(pl.UInt8, DType.UINT8)

    assert_conversion(pl.Float64, DType.FLOAT64)
    assert_conversion(pl.Float32, DType.FLOAT32)

    assert_conversion(pl.Utf8, DType.STRING)
    assert_conversion(pl.Boolean, DType.BOOLEAN)

    assert_conversion(pl.Date, DType.DATE)
    assert_conversion(pl.Time, DType.TIME)
    assert_conversion(pl.Datetime, DType.DATETIME)
    assert_conversion(pl.Datetime("ms"), DType.DATETIME)
    assert_conversion(pl.Datetime("us"), DType.DATETIME)
    assert_conversion(pl.Datetime("ns"), DType.DATETIME)


def test_dtype_to_polars():
    def assert_conversion(type_: DType, expected):
        assert type_.to_polars() == expected

    assert_conversion(DType.INT64, pl.Int64)
    assert_conversion(DType.INT32, pl.Int32)
    assert_conversion(DType.INT16, pl.Int16)
    assert_conversion(DType.INT8, pl.Int8)

    assert_conversion(DType.UINT64, pl.UInt64)
    assert_conversion(DType.UINT32, pl.UInt32)
    assert_conversion(DType.UINT16, pl.UInt16)
    assert_conversion(DType.UINT8, pl.UInt8)

    assert_conversion(DType.FLOAT64, pl.Float64)
    assert_conversion(DType.FLOAT32, pl.Float32)

    assert_conversion(DType.STRING, pl.Utf8)
    assert_conversion(DType.BOOLEAN, pl.Boolean)

    assert_conversion(DType.DATE, pl.Date)
    assert_conversion(DType.TIME, pl.Time)
    assert_conversion(DType.DATETIME, pl.Datetime("us"))



from __future__ import annotations

import pytest

from example.run_pipeline import main as example_flow_main
from example.simple_pipeline import main as simple_pipeline_main
from example.visualization import main as visualization_main
from example_imperative.run_pipeline import main as example_imperative_flow_main
from example_interactive.run_tasks_interactively import main as example_interactive_main
from example_postgres.run_pipeline import main as example_postgres_flow_main


@pytest.mark.parametrize(
    "fn",
    [
        example_flow_main,
        simple_pipeline_main,
        visualization_main,
        example_imperative_flow_main,
        example_postgres_flow_main,
        example_interactive_main,
    ],
)
def test_examples(fn):
    """
    This test just runs the example pipeline that we provide in example/run_pipeline.py
    """

    fn()



from __future__ import annotations

import pandas as pd
import pytest
import sqlalchemy as sa

from pydiverse.pipedag import Blob, ConfigContext, Flow, Stage, Table
from pydiverse.pipedag.container import RawSql
from pydiverse.pipedag.context import StageLockContext
from pydiverse.pipedag.context.context import CacheValidationMode
from pydiverse.pipedag.materialize.core import (
    AUTO_VERSION,
    input_stage_versions,
    materialize,
)

# Parameterize all tests in this file with several instance_id configurations
from tests.fixtures.instances import ALL_INSTANCES, skip_instances, with_instances
from tests.util import compile_sql, select_as
from tests.util import tasks_library as m
from tests.util import tasks_library_imperative as m2
from tests.util.spy import spy_task
from tests.util.tasks_library import get_task_logger

try:
    import polars as pl
except ImportError:
    pl = None

# snowflake tests are too slow, possibly they could move to nightly tests
pytestmark = [with_instances(tuple(set(ALL_INSTANCES) - {"snowflake"}))]


# Test Basic Cache Invalidation Behaviour


def test_change_bound_argument(mocker):
    input_list = [1]

    with Flow() as flow:
        with Stage("stage_1"):
            out = m.noop(input_list)
            child = m.noop2(out)

    # Initial Call
    with StageLockContext():
        result = flow.run()
        assert result.get(out)[0] == 1
        assert result.get(child)[0] == 1

    # Calling flow.run again shouldn't call the task
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    with StageLockContext():
        result = flow.run()
        assert result.get(out)[0] == 1
        assert result.get(child)[0] == 1
        out_spy.assert_not_called()
        child_spy.assert_not_called()

    # Changing the input object should invalidate the cache
    input_list[0] = 2
    with StageLockContext():
        result = flow.run()
        assert result.get(out)[0] == 2
        assert result.get(child)[0] == 2
        out_spy.assert_called_once()
        child_spy.assert_called_once()


def test_changed_cache_fn_literal(mocker):
    cache_value = 0

    def cache():
        return cache_value

    @materialize(cache=cache, version="1.0")
    def return_cache_value():
        return cache_value

    with Flow() as flow:
        with Stage("stage_1"):
            out = return_cache_value()
            child = m.noop(out)

    # Initial Call
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 0
        assert result.get(child) == 0

    # Calling flow.run again shouldn't call the task
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 0
        assert result.get(child) == 0
        out_spy.assert_not_called()
        child_spy.assert_not_called()

    # Changing the cache value should cause it to get called again
    cache_value = 1
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 1
        assert result.get(child) == 1
        out_spy.assert_called_once()
        child_spy.assert_called_once()


@pytest.mark.parametrize("imperative", [False, True])
def test_change_cache_fn_table(mocker, imperative):
    cache_value = 0

    def cache():
        return cache_value

    @materialize(cache=cache, version="1.0")
    def return_cache_table():
        tbl = Table(select_as(cache_value, "x"))
        return tbl.materialize() if imperative else tbl

    @materialize(input_type=pd.DataFrame, version="1.0")
    def get_first(table, col):
        return int(table[col][0])

    with Flow() as flow:
        with Stage("stage_1"):
            out = return_cache_table()
            child = get_first(out, "x")

    # Initial Call
    with StageLockContext():
        result = flow.run()
        assert result.get(child) == 0

    # Calling flow.run again shouldn't call the task
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    with StageLockContext():
        result = flow.run()
        assert result.get(child) == 0
        out_spy.assert_not_called()
        child_spy.assert_not_called()

    # Changing the cache value should cause it to get called again
    cache_value = 1
    with StageLockContext():
        result = flow.run()
        assert result.get(child) == 1
        out_spy.assert_called_once()
        child_spy.assert_called_once()


@with_instances("postgres")
def test_change_cache_fn_blob(mocker):
    cache_value = 0

    def cache():
        return cache_value

    @materialize(cache=cache, version="1.0")
    def return_cache_blob():
        return Blob(cache_value)

    with Flow() as flow:
        with Stage("stage_1"):
            out = return_cache_blob()
            child = m.noop(out)

    # Initial Call
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 0
        assert result.get(child) == 0

    # Calling flow.run again shouldn't call the task
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 0
        assert result.get(child) == 0
        out_spy.assert_not_called()
        child_spy.assert_not_called()

    # Changing the cache value should cause it to get called again
    cache_value = 1
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 1
        assert result.get(child) == 1
        out_spy.assert_called_once()
        child_spy.assert_called_once()


def test_change_task_version_literal(mocker):
    with Flow() as flow:
        with Stage("stage_1"):
            out = m.one()
            child = m.noop(out)

    # Initial Call
    out._version = "VERSION 0"
    assert flow.run().successful

    # Second Call (Should be cached)
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    assert flow.run().successful
    out_spy.assert_not_called()
    child_spy.assert_not_called()

    # Changing the version should invalidate the cache, but the child still
    # shouldn't get called because the parent task still returned the same value.
    out._version = "VERSION 1"
    assert flow.run().successful
    out_spy.assert_called_once()
    child_spy.assert_not_called()


def test_change_task_version_table(mocker):
    with Flow() as flow:
        with Stage("stage_1"):
            out = m.simple_dataframe()
            child = m.noop(out)
            child2 = m.noop_sql(out)  # lazy=False task
            child_lazy = m.noop_lazy(out)  # lazy=True task

    # Initial Call
    out._version = "VERSION 0"
    assert flow.run().successful

    # Second Call (Should be cached)
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    child2_spy = spy_task(mocker, child2)
    child_lazy_spy = spy_task(mocker, child_lazy)
    assert flow.run().successful
    out_spy.assert_not_called()
    child_spy.assert_not_called()
    child2_spy.assert_not_called()
    child_lazy_spy.assert_called_once()

    # Changing the version should invalidate the cache. This should also invalidate
    # the child task because it receives the table as input.
    out._version = "VERSION 1"
    assert flow.run().successful
    out_spy.assert_called_once()
    child_spy.assert_called_once()
    child2_spy.assert_called_once()
    child_lazy_spy.assert_called_once()


@with_instances("postgres")
def test_change_task_version_blob(mocker):
    with Flow() as flow:
        with Stage("stage_1"):
            out = m.object_blob({"x": "y"})
            child = m.as_blob(out)

    # Initial Call
    out._version = "VERSION 0"
    assert flow.run().successful

    # Second Call (Should be cached)
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    assert flow.run().successful
    out_spy.assert_not_called()
    child_spy.assert_not_called()

    # Changing the version should invalidate the cache. This should also invalidate
    # the child task because it receives the blob as input.
    out._version = "VERSION 1"
    assert flow.run().successful
    out_spy.assert_called_once()
    child_spy.assert_called_once()


def test_change_lazy_query(mocker):
    query_value = 1

    @materialize(lazy=True, nout=2)
    def lazy_task():
        return 0, Table(select_as(query_value, "x"), name="lazy_table")

    @materialize(input_type=pd.DataFrame, version="1.0")
    def get_first(table, col):
        return int(table[col][0])

    with Flow() as flow:
        with Stage("stage_1"):
            const, lazy = lazy_task()
            value = get_first(lazy, "x")
            const2 = m.noop(const)
            lazy2 = m.noop_sql(lazy)  # lazy=False task
            lazy3 = m.noop_lazy(lazy)

    # Initial Run
    lazy_spy = spy_task(mocker, lazy)
    with StageLockContext():
        result = flow.run()
        assert result.successful
        assert result.get(value) == 1
        lazy_spy.assert_called_once()

    # Second run, because the task is lazy, it should always get called.
    # The value task however shouldn't get called.
    value_spy = spy_task(mocker, value)
    const2_spy = spy_task(mocker, const2)
    lazy2_spy = spy_task(mocker, lazy2)
    lazy3_spy = spy_task(mocker, lazy3)
    with StageLockContext():
        result = flow.run()
        assert result.successful
        assert result.get(value) == 1
        lazy_spy.assert_called_once()
        value_spy.assert_not_called()
        const2_spy.assert_not_called()
        lazy2_spy.assert_not_called()
        lazy3_spy.assert_called_once()

    # Third run with changed query_value
    query_value = 2
    with StageLockContext():
        result = flow.run()
        assert result.successful
        assert result.get(value) == 2
        lazy_spy.assert_called_once()
        value_spy.assert_called_once()
        const2_spy.assert_not_called()
        lazy2_spy.assert_called_once()
        lazy3_spy.assert_called_once()


def test_change_raw_sql(mocker):
    query_value = 1

    @materialize(lazy=True, nout=2, version="1.0")
    def raw_task(stage):
        raw_sql = compile_sql(select_as(query_value, "x"))
        return 0, RawSql(raw_sql, "raw_task")

    @materialize(version="1.0")
    def raw_child(raw):
        return raw.sql

    with Flow() as flow:
        with Stage("stage_1") as s:
            const, raw = raw_task(s)
            child = raw_child(raw)
            const = m.noop(const)

    # Initial Run
    raw_spy = spy_task(mocker, raw)
    with StageLockContext():
        result = flow.run()
        assert result.successful
        raw_spy.assert_called_once()

    # Second run, because the task is lazy, it should always get called.
    # The value task however shouldn't get called.
    child_spy = spy_task(mocker, child)
    const_spy = spy_task(mocker, const)
    with StageLockContext():
        result = flow.run()
        assert result.successful
        assert "SELECT 1" in result.get(child).upper()
        raw_spy.assert_called_once()
        child_spy.assert_not_called()
        const_spy.assert_not_called()

    # Third run with changed query_str
    query_value = 2
    with StageLockContext():
        result = flow.run()
        assert result.successful
        assert "SELECT 2" in result.get(child).upper()
        raw_spy.assert_called_once()
        child_spy.assert_called_once()
        const_spy.assert_not_called()


def test_change_task_stage_literal(mocker):
    with Flow() as flow:
        with Stage("stage_1"):
            one = m.one()
        with Stage("stage_2"):
            m.noop(0)  # This is to clear the stage
        with Stage("stage_3"):
            child = m.noop(one)
    _ = child

    assert flow.run().successful

    with Flow() as flow:
        with Stage("stage_2"):
            one = m.one()
        with Stage("stage_3"):
            child = m.noop(one)

    one_spy = spy_task(mocker, one)
    child_spy = spy_task(mocker, child)

    # Moving the one task to a different stage should cause it to be called again,
    # but because its return value is the same, the child task shouldn't get called.
    assert flow.run().successful
    one_spy.assert_called_once()
    child_spy.assert_not_called()


def test_change_task_stage_table(mocker):
    with Flow() as flow:
        with Stage("stage_1"):
            table = m.simple_dataframe()
        with Stage("stage_2"):
            m.noop(0)  # This is to clear the stage
        with Stage("stage_3"):
            child = m.noop(table)

    assert flow.run().successful

    with Flow() as flow:
        with Stage("stage_2"):
            table = m.simple_dataframe()
        with Stage("stage_3"):
            child = m.noop(table)

    table_spy = spy_task(mocker, table)
    child_spy = spy_task(mocker, child)

    # Moving the table task to a different stage should cause it to be called again,
    # and because it returns a table, any child task should also get invalidated.
    assert flow.run().successful
    table_spy.assert_called_once()
    child_spy.assert_called_once()


@with_instances("postgres")
def test_change_task_stage_blob(mocker):
    with Flow() as flow:
        with Stage("stage_1"):
            blob = m.as_blob(1)
        with Stage("stage_2"):
            m.noop(0)  # This is to clear the stage
        with Stage("stage_3"):
            child = m.noop(blob)

    assert flow.run().successful

    with Flow() as flow:
        with Stage("stage_2"):
            blob = m.as_blob(1)
        with Stage("stage_3"):
            child = m.noop(blob)

    blob_spy = spy_task(mocker, blob)
    child_spy = spy_task(mocker, child)

    # Moving the blob task to a different stage should cause it to be called again,
    # and because it returns a blob, any child task should also get invalidated.
    assert flow.run().successful
    blob_spy.assert_called_once()
    child_spy.assert_called_once()


def test_different_task_same_input(mocker):
    with Flow() as flow:
        with Stage("stage_1"):
            one = m.one()
            two = m.two()

    # Initial run
    with StageLockContext():
        result = flow.run()
        assert result.successful
        assert result.get(one) == 1
        assert result.get(two) == 2

    # Second Run should be cached
    one_spy = spy_task(mocker, one)
    two_spy = spy_task(mocker, two)

    with StageLockContext():
        result = flow.run()
        assert result.successful
        assert result.get(one) == 1
        assert result.get(two) == 2
        one_spy.assert_not_called()
        two_spy.assert_not_called()


def test_same_task_different_stages(mocker):
    with Flow() as flow:
        with Stage("stage_1"):
            out_s1 = m.one()
        with Stage("stage_2"):
            out_s2 = m.one()

    # Initial run
    with StageLockContext():
        result = flow.run()
        assert result.successful
        assert result.get(out_s1) == 1
        assert result.get(out_s2) == 1

    # Second Run should be cached
    out_spy = spy_task(mocker, out_s1)

    with StageLockContext():
        result = flow.run()
        assert result.successful
        assert result.get(out_s1) == 1
        assert result.get(out_s2) == 1
        out_spy.assert_not_called()


def test_change_version_table(mocker):
    cache_value = 0

    def cache():
        return cache_value

    @materialize(lazy=True, cache=cache)
    def return_cache_table():
        return Table(select_as(cache_value, "x"))

    def get_flow(version):
        @materialize(version=version, input_type=pd.DataFrame)
        def named_copy(table: pd.DataFrame):
            return Table(table, "_table_copy")

        with Flow() as flow:
            with Stage("stage_1"):
                out = return_cache_table()
                cpy = named_copy(out)
        return flow, out, cpy

    flow, out, cpy = get_flow(version="1.0")
    # Initial Call
    with StageLockContext():
        result = flow.run()
        assert result.get(cpy)["x"].iloc[0] == 0

    # Calling flow.run again shouldn't call the task
    out_spy = spy_task(mocker, out)
    cpy_spy = spy_task(mocker, cpy)
    with StageLockContext():
        result = flow.run()
        assert result.get(cpy)["x"].iloc[0] == 0
        out_spy.assert_called_once()  # lazy task is always called
        cpy_spy.assert_not_called()

    # Changing the cache value should cause it to get called again
    cache_value = 1
    with StageLockContext():
        result = flow.run()
        assert result.get(cpy)["x"].iloc[0] == 1
        out_spy.assert_called_once()
        cpy_spy.assert_called_once()

    # recreating flow without version change should not invalidate cache
    flow, out, cpy = get_flow(version="1.0")
    out_spy = spy_task(mocker, out)
    cpy_spy = spy_task(mocker, cpy)
    with StageLockContext():
        result = flow.run()
        assert result.get(cpy)["x"].iloc[0] == 1
        out_spy.assert_called_once()  # lazy task is always called
        cpy_spy.assert_not_called()

    # Changing the version value should cause cpy to get recreated again
    flow, out, cpy = get_flow(version="1.1")
    out_spy = spy_task(mocker, out)
    cpy_spy = spy_task(mocker, cpy)
    with StageLockContext():
        result = flow.run()
        assert result.get(cpy)["x"].iloc[0] == 1
        out_spy.assert_called_once()  # lazy task is always called
        cpy_spy.assert_called_once()

    # Changing the version value should cause cpy to get recreated again
    for _ in range(3):
        flow, out, cpy = get_flow(version=None)
        out_spy = spy_task(mocker, out)
        cpy_spy = spy_task(mocker, cpy)
        with StageLockContext():
            result = flow.run()
            assert result.get(cpy)["x"].iloc[0] == 1
            out_spy.assert_called_once()  # lazy task is always called
            cpy_spy.assert_called_once()


@pytest.mark.parametrize("ignore_task_version", [True, False])
@pytest.mark.parametrize("disable_cache_function", [True, False])
@pytest.mark.parametrize("mode", ["ASSERT_NO_FRESH_INPUT", "NORMAL"])
def test_cache_validation_mode_assert(
    ignore_task_version, disable_cache_function, mode
):
    mode = getattr(CacheValidationMode, mode.upper())
    kwargs = dict(
        ignore_task_version=ignore_task_version,
        disable_cache_function=disable_cache_function,
        cache_validation_mode=mode,
    )
    if not disable_cache_function and mode == CacheValidationMode.NORMAL:
        pytest.skip("Combination does not provoke exception")
        return
    error = (
        AssertionError
        if mode == CacheValidationMode.ASSERT_NO_FRESH_INPUT
        else ValueError
    )

    cache_value = 0
    return_value = 0

    def cache():
        return cache_value

    def get_flow(version):
        @materialize(version=version, cache=cache)
        def source():
            return pd.DataFrame(dict(x=[return_value]))

        @materialize(lazy=True, cache=cache)
        def lazy_source():
            return Table(select_as(return_value, "x"))

        with Flow() as flow:
            with Stage("stage_1"):
                s1 = source()
                s2 = lazy_source()
        return flow, s1, s2

    flow, s1, s2 = get_flow(version="1.0")
    # Initial Call
    with StageLockContext():
        result = flow.run(
            cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID,
            disable_cache_function=False,
            ignore_task_version=False,
        )
        assert result.successful
        assert result.get(s1, as_type=pd.DataFrame)["x"].iloc[0] == return_value
        assert result.get(s2, as_type=pd.DataFrame)["x"].iloc[0] == return_value

    if disable_cache_function or ignore_task_version:
        with pytest.raises(error):
            result = flow.run(**kwargs)
            assert result.successful

    cache_value = 1  # IGNORE_FRESH_INPUT should be implied
    if (
        not disable_cache_function
        and not ignore_task_version
        and mode != CacheValidationMode.NORMAL
    ):
        result = flow.run(
            cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID,
            disable_cache_function=False,
            ignore_task_version=False,
        )
        assert result.successful

    return_value = 1
    with pytest.raises(error):
        flow.run(**kwargs)

    return_value = 0
    flow, _, _ = get_flow(version="1.1")
    with pytest.raises(error):
        flow.run(**kwargs)


@with_instances("postgres", "local_table_cache")
@pytest.mark.parametrize("ignore_task_version", [True, False])
@pytest.mark.parametrize("disable_cache_function", [True, False])
@pytest.mark.parametrize(
    "mode", ["NORMAL", "IGNORE_FRESH_INPUT", "FORCE_FRESH_INPUT", "FORCE_CACHE_INVALID"]
)
@pytest.mark.parametrize("imperative", [False, True])
def test_cache_validation_mode(
    ignore_task_version, disable_cache_function, mode, imperative, mocker
):
    _m = m2 if imperative else m
    mode = getattr(CacheValidationMode, mode.upper())
    if disable_cache_function and mode == CacheValidationMode.NORMAL:
        pytest.skip("Cannot disable cache function in mode NORMAL")
        return

    kwargs = dict(
        ignore_task_version=ignore_task_version,
        disable_cache_function=disable_cache_function,
        cache_validation_mode=mode,
    )
    cache_calls = [0]  # consider using mocker also for this
    cache_value = 0

    def cache():
        cache_calls[0] += 1
        return cache_value

    def cache2(df):
        cache_calls[0] += 1
        return cache_value

    @materialize(lazy=True, cache=cache)
    def return_cache_table_lazy():
        tbl = Table(select_as(cache_value, "x"))
        return tbl.materialize() if imperative else tbl

    @materialize(version=None, cache=cache)
    def return_cache_table_always():
        df = pd.DataFrame(dict(x=[cache_value]))
        return Table(df).materialize() if imperative else df

    @materialize(version=AUTO_VERSION, cache=cache2, input_type=pd.DataFrame)
    def return_cache_table_auto(df: pd.DataFrame):
        return Table(df).materialize() if imperative else df

    @materialize(version="1.0", cache=cache)
    def return_cache_table():
        df = pd.DataFrame(dict(x=[cache_value]))
        return Table(df).materialize() if imperative else df

    @input_stage_versions(lazy=True, input_type=sa.Table)
    def dummy_copy_inputs(
        transaction: dict[str, sa.sql.expressions.Alias],
        other: dict[str, sa.sql.expressions.Alias],
    ):
        _ = other  # we cannot make any assumptions on the other stage version
        assert len(transaction) == 0
        return Table(sa.select(sa.literal(1).label("a")), name="dummy")

    @input_stage_versions(lazy=True, input_type=sa.Table)
    def validate_stage(
        transaction: dict[str, sa.sql.expressions.Alias],
        other: dict[str, sa.sql.expressions.Alias],
    ):
        _ = other  # we cannot make any assumptions on the other stage version
        get_task_logger().info(f"Transaction tables: {transaction}")
        transaction_schema = {tbl.original.schema for tbl in transaction.values()} - {
            tbl.original.schema for tbl in other.values()
        }
        assert transaction_schema
        assert transaction_schema.pop().startswith(
            {tbl.original.schema for tbl in other.values()}.pop().split("__")[0]
        )
        assert (
            len([tbl for tbl in transaction.keys() if not tbl.endswith("__copy")]) == 12
        )

    @input_stage_versions(lazy=True, input_type=sa.Table)
    def validate_stage2(
        transaction: dict[str, sa.sql.expressions.Alias],
        other: dict[str, sa.sql.expressions.Alias],
    ):
        # it is expected that we have a "Failed to retrieve"-exception in other stage
        _ = other
        assert len(transaction) == 1

    def get_flow():
        with Flow() as flow:
            with Stage("stage_1"):
                x = dummy_copy_inputs()
                _ = m.noop(x)
                out_lazy = return_cache_table_lazy()
                out_always = return_cache_table_always()
                out_auto = return_cache_table_auto(out_lazy)
                out = return_cache_table()
                outs = [out_lazy, out_always, out_auto, out]
                cpy = [_m.noop(o) for o in outs]
                ind1 = _m.simple_dataframe()
                ind2 = _m.simple_lazy_table()
                validate_stage()
                validate_stage2(ind2)
        return flow, outs, cpy, [ind1, ind2]

    flow, outs, cpy, ind = get_flow()
    # Initial Call
    with StageLockContext():
        result = flow.run(
            cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID,
            disable_cache_function=False,
            ignore_task_version=False,
        )
        assert all(result.get(c, as_type=pd.DataFrame)["x"].iloc[0] == 0 for c in cpy)
    assert cache_calls[0] == 4
    cache_calls[0] = 0

    # Calling flow.run again
    out_spy = [spy_task(mocker, o) for o in outs]
    cpy_spy = [spy_task(mocker, c) for c in cpy]
    ind_spy = [spy_task(mocker, i) for i in ind]
    with StageLockContext():
        result = flow.run(**kwargs)
        assert all(result.get(c, as_type=pd.DataFrame)["x"].iloc[0] == 0 for c in cpy)
        out_spy[0].assert_called_once()  # lazy task is always called
        out_spy[1].assert_called_once()  # version=None task is always called
        if (
            mode
            in [
                CacheValidationMode.FORCE_FRESH_INPUT,
                CacheValidationMode.FORCE_CACHE_INVALID,
            ]
            and not disable_cache_function
        ):
            out_spy[2].assert_called(2)
        else:
            out_spy[2].assert_called_once()  # version=AUTO_VERSION task is called once
        if (
            mode
            not in [
                CacheValidationMode.FORCE_FRESH_INPUT,
                CacheValidationMode.FORCE_CACHE_INVALID,
            ]
            and not ignore_task_version
        ):
            out_spy[3].assert_not_called()
        else:
            out_spy[3].assert_called_once()
        if mode not in [CacheValidationMode.FORCE_CACHE_INVALID]:
            if ignore_task_version:
                all(spy.assert_called_once() for spy in cpy_spy)
            elif disable_cache_function:
                if mode == CacheValidationMode.IGNORE_FRESH_INPUT:
                    cpy_spy[0].assert_not_called()  # lazy cache_fn_hash is loaded
                    cpy_spy[1].assert_called_once()  # version=None is always called
                    cpy_spy[2].assert_not_called()  # cache_key is loaded
                    cpy_spy[3].assert_not_called()  # cache_key is loaded
                else:
                    all(spy.assert_called_once() for spy in cpy_spy)
            else:
                cpy_spy[0].assert_not_called()
                cpy_spy[1].assert_called_once()  # version=None is always called
                cpy_spy[2].assert_not_called()  # cache_fn_hash is stable
                cpy_spy[3].assert_not_called()
            if ignore_task_version:
                ind_spy[0].assert_called_once()
            else:
                ind_spy[0].assert_not_called()
            ind_spy[1].assert_called_once()  # lazy is always called
        else:
            all(spy.assert_called_once() for spy in cpy_spy)
            all(spy.assert_called_once() for spy in ind_spy)
    if disable_cache_function:
        assert cache_calls[0] == 0
    else:
        assert cache_calls[0] == 4

    # Changing the cache value should cause it to get called again
    cache_value = 1
    with StageLockContext():
        result = flow.run(**kwargs)
        if mode == CacheValidationMode.IGNORE_FRESH_INPUT and not ignore_task_version:
            assert all(
                result.get(c, as_type=pd.DataFrame)["x"].iloc[0] == res
                for c, res in zip(cpy, [cache_value, cache_value, cache_value, 0])
            )
        else:
            assert all(
                result.get(c, as_type=pd.DataFrame)["x"].iloc[0] == cache_value
                for c in cpy
            )
        out_spy[0].assert_called_once()  # lazy task is always called
        out_spy[1].assert_called_once()  # version=None task is always called
        if disable_cache_function and mode in [
            CacheValidationMode.FORCE_FRESH_INPUT,
            CacheValidationMode.FORCE_CACHE_INVALID,
        ]:
            out_spy[2].assert_called_once()
        else:
            out_spy[2].assert_called(2)  # version=AUTO_VERSION task is called twice
        if mode == CacheValidationMode.IGNORE_FRESH_INPUT and not ignore_task_version:
            out_spy[3].assert_not_called()
        else:
            out_spy[3].assert_called_once()
        if mode not in [CacheValidationMode.FORCE_CACHE_INVALID]:
            all(spy.assert_called_once() for spy in cpy_spy)
            if ignore_task_version:
                ind_spy[0].assert_called_once()
            else:
                ind_spy[0].assert_not_called()
            ind_spy[1].assert_called_once()  # lazy is always called
        else:
            all(spy.assert_called_once() for spy in cpy_spy)
            all(spy.assert_called_once() for spy in ind_spy)


@skip_instances("postgres", "local_table_cache", "snowflake")
@pytest.mark.parametrize("ignore_task_version", [False])
@pytest.mark.parametrize("disable_cache_function", [False])
@pytest.mark.parametrize(
    "mode", ["NORMAL", "IGNORE_FRESH_INPUT", "FORCE_FRESH_INPUT", "FORCE_CACHE_INVALID"]
)
@pytest.mark.parametrize("imperative", [True])
def test_cache_validation_mode_reduced(
    ignore_task_version, disable_cache_function, mode, imperative, mocker
):
    # Reduce combinatorial space for duckdb to avoid timeout after 10min
    # duckdb is particularly slow for those tests (~10x: 7-15s instead of 1-2s).
    # This reduced combinatorial space is even too much for snowflake. There,
    # we have 1-2 seconds roundtrip time for many individual database requests.
    # And even just reflecting a SQLAlchemy tables issues several of those.
    # Thus this test takes >30min alone for snowflake. In realistic scenarios,
    # the 1-2s per request should not matter too much, though.
    test_cache_validation_mode(
        ignore_task_version, disable_cache_function, mode, imperative, mocker
    )


@pytest.mark.parametrize("n", [1, 2, 15])
def test_partial_stage_cache_valid(mocker, n):
    cache_value = 0

    def cache():
        return cache_value

    @materialize(lazy=True, cache=cache)
    def return_cache_table():
        return Table(select_as(cache_value, "x"))

    @materialize(version="1.0")
    def n_tables(n):
        return [pd.DataFrame(dict(x=[i])) for i in range(n)]

    @materialize(lazy=True)
    def n_lazy_tables(n):
        return [Table(select_as(i, "x")) for i in range(n)]

    def get_flow(version):
        @materialize(version=version, input_type=pd.DataFrame)
        def named_copy(table: pd.DataFrame):
            return Table(table, "_table_copy")

        with Flow() as flow:
            with Stage("stage_1"):
                # these tasks are cache valid in all but the first call:
                tbls = n_tables(n)
                lazy_tbls = n_lazy_tables(n)
                # these tables may be cache invalidated by cache_value
                out = return_cache_table()
                cpy = named_copy(out)
        return flow, out, cpy, tbls, lazy_tbls

    flow, out, cpy, tbls, lazy_tbls = get_flow(version="1.0")
    # Initial Call
    with StageLockContext():
        result = flow.run()
        assert result.get(cpy)["x"].iloc[0] == 0

    # Calling flow.run again shouldn't call the task
    out_spy = spy_task(mocker, out)
    cpy_spy = spy_task(mocker, cpy)
    tbls_spy = spy_task(mocker, tbls)
    lazy_tbls_spy = spy_task(mocker, lazy_tbls)
    with StageLockContext():
        result = flow.run()
        assert result.get(cpy)["x"].iloc[0] == 0
        out_spy.assert_called_once()  # lazy task is always called
        cpy_spy.assert_not_called()
        tbls_spy.assert_not_called()
        lazy_tbls_spy.assert_called_once()  # lazy task is always called

    # Changing the cache value should cause it to get called again
    # In this case tbls and lazy_tbls are still cache valid and thus need to be
    # copied over to transaction schema
    cache_value = 1
    with StageLockContext():
        result = flow.run()
        assert result.get(cpy)["x"].iloc[0] == 1
        out_spy.assert_called_once()
        cpy_spy.assert_called_once()
        tbls_spy.assert_not_called()
        lazy_tbls_spy.assert_called_once()  # lazy task is always called


def test_ignore_task_version(mocker):
    cfg = ConfigContext.get().evolve(cache_validation=dict(ignore_task_version=True))

    with Flow() as flow:
        with Stage("stage_1"):
            out = m.noop([1])
            child = m.noop2(out)

    # Initial Call
    with StageLockContext():
        result = flow.run(config=cfg)
        assert result.get(out)[0] == 1
        assert result.get(child)[0] == 1

    # Calling flow.run again should still call the task (disabled caching)
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    for _ in range(3):
        with StageLockContext():
            result = flow.run(config=cfg)
            assert result.get(out)[0] == 1
            assert result.get(child)[0] == 1
            out_spy.assert_called_once()
            child_spy.assert_called_once()


@pytest.mark.polars
def test_lazy_table_without_query_string(mocker):
    value = None

    @materialize(lazy=True, nout=5)
    def falsely_lazy_task():
        return (
            Table(pd.DataFrame({"x": [value]}), name="pd_table"),
            Table(pl.DataFrame({"x": [value]}), name="pl_table"),
            Table(pl.DataFrame({"x": [value]}).lazy(), name="pl_lazy_table"),
            select_as(2, "y"),
            3,
        )

    def get_flow():
        with Flow() as flow:
            with Stage("stage_1"):
                pd_tbl, pl_tbl, pl_lazy_tbl, select_tbl, constant = falsely_lazy_task()
                res_pd = m.take_first(pd_tbl, as_int=True)
                res_pl = m.take_first(pl_tbl, as_int=True)
                res_pl_lazy = m.take_first(pl_lazy_tbl, as_int=True)
                res_select = m.noop(select_tbl)
                res_constant = m.noop(constant)
        return flow, res_pd, res_pl, res_pl_lazy, res_select, res_constant

    value = 0
    flow, res_pd, res_pl, res_pl_lazy, res_select, res_constant = get_flow()
    with StageLockContext():
        result = flow.run()
        assert result.get(res_pd) == 0
        assert result.get(res_pl) == 0
        assert result.get(res_pl_lazy) == 0

    value = 1
    flow, res_pd, res_pl, res_pl_lazy, res_select, res_constant = get_flow()
    res_pd_spy = spy_task(mocker, res_pd)
    res_pl_spy = spy_task(mocker, res_pl)
    res_pl_lazy_spy = spy_task(mocker, res_pl_lazy)
    select_spy = spy_task(mocker, res_select)
    constant_spy = spy_task(mocker, res_constant)
    with StageLockContext():
        result = flow.run()
        assert result.get(res_pd) == 1
        assert result.get(res_pl) == 1
        assert result.get(res_pl_lazy) == 1
        # res_pd is downstream of a pd.DataFrame from a lazy task,
        # which should always be cache invalid. Hence, it should always be called.
        res_pd_spy.assert_called_once()

        # res_pd is downstream of a pl.DataFrame from a lazy task,
        # which should always be cache invalid. Hence, it should always be called.
        res_pl_spy.assert_called_once()

        # res_pd is downstream of a pl.LazyFrame from a lazy task,
        # which should always be cache invalid. Hence, it should always be called.
        # To avoid cache-invalidating the LazyFrame, we should use AUTOVERSION.
        res_pl_lazy_spy.assert_called_once()

        # res_select is downstream of an SQL query `select_tbl` (which did not change)
        # from a lazy task, hence select_tbl should be cache valid and the task
        # producing res_select should not be called.
        select_spy.assert_not_called()

        # res_constant is downstream of the constant `constant` (which did not change)
        # from a lazy task, hence res_constant should be cache valid and the task
        # producing res_select should not be called.
        constant_spy.assert_not_called()



from __future__ import annotations

import copy

import pandas as pd
import pytest
import sqlalchemy as sa

from pydiverse.pipedag import ConfigContext, Flow, Stage, StageLockContext, Table
from pydiverse.pipedag.container import Schema
from pydiverse.pipedag.core.config import PipedagConfig, create_basic_pipedag_config
from tests.fixtures.instances import with_instances
from tests.util import tasks_library as m
from tests.util import tasks_library_imperative as m2


@with_instances(
    "postgres",
    "local_table_cache",
    "local_table_cache_inout",
    "local_table_cache_inout_numpy",
)
@pytest.mark.parametrize("imperative", [False, True])
def test_get_output_from_store(imperative):
    _m = m2 if imperative else m
    with Flow() as f:
        with Stage("stage_1"):
            df1 = _m.pd_dataframe({"x": [0, 1, 2, 3]})
            df2 = _m.pd_dataframe({"y": [0, 1, 2, 3]})
            dataframes = _m.create_tuple(df1, df2)

    # We only use the StageLockContext for testing
    with StageLockContext():
        result = f.run()

        # Call on MaterializingTask
        pd.testing.assert_frame_equal(
            df1.get_output_from_store(as_type=pd.DataFrame),
            result.get(df1, as_type=pd.DataFrame),
        )

        pd.testing.assert_frame_equal(
            df2.get_output_from_store(as_type=pd.DataFrame),
            result.get(df2, as_type=pd.DataFrame),
        )

        pd.testing.assert_frame_equal(
            dataframes.get_output_from_store(as_type=pd.DataFrame)[1],
            result.get(dataframes, as_type=pd.DataFrame)[1],
        )

        # since df1 and df2 are outputs from tasks with the same name (pd_dataframe)
        # ignoring the position hash when retrieving
        # should just return the output from the latest call
        pd.testing.assert_frame_equal(
            df1.get_output_from_store(
                as_type=pd.DataFrame, ignore_position_hashes=True
            ),
            result.get(df2, as_type=pd.DataFrame),
        )

        # Call on MaterializingTaskGetItem
        pd.testing.assert_frame_equal(
            dataframes[0].get_output_from_store(as_type=pd.DataFrame),
            dataframes.get_output_from_store(as_type=pd.DataFrame)[0],
        )

        pd.testing.assert_frame_equal(
            dataframes[0].get_output_from_store(
                as_type=pd.DataFrame, ignore_position_hashes=True
            ),
            dataframes.get_output_from_store(
                as_type=pd.DataFrame, ignore_position_hashes=True
            )[0],
        )


@with_instances("postgres")
@pytest.mark.parametrize("imperative", [False, True])
def test_call_task_outside_flow(imperative):
    _m = m2 if imperative else m
    # Literal Values
    assert m.one() == 1
    assert m.create_tuple(m.one(), m.two()) == (1, 2)
    assert m.noop("foo") == "foo"

    # Table
    df1 = _m.pd_dataframe({"x": [0, 1, 2, 3]})
    df2 = m.noop(df1)
    expected = pd.DataFrame({"x": [0, 1, 2, 3]})

    for df in [df1, df2]:
        pd.testing.assert_frame_equal(df, expected)
        _m.assert_table_equal(df, expected)

    # Lazy Table
    expr = _m.simple_lazy_table()
    expected_expr = (
        "SELECT 0 AS col1, '0' AS col2 "
        "UNION ALL SELECT 1 AS col1, '1' AS col2 "
        "UNION ALL SELECT 2 AS col1, '2' AS col2 "
        "UNION ALL SELECT 3 AS col1, '3' AS col2"
    )
    config_context = ConfigContext.get()
    assert (
        str(
            expr.compile(
                config_context.store.table_store.engine,
                compile_kwargs={"literal_binds": True},
            )
        )
        == expected_expr
    )

    tbl = m.noop(sa.Table("foo", sa.MetaData()).alias("bar"))
    assert tbl.name == "bar"
    assert tbl.original.name == "foo"

    # Blob
    expected_blob = {"a": {"b": 1, "c": 2}, "d": [1, (2, 3)]}
    blob = m.as_blob(expected_blob)

    assert blob is expected_blob
    m.assert_blob_equal(blob, expected_blob)

    # RawSql
    expected_query = "SELECT 1 as x"
    raw_sql = m.as_raw_sql(expected_query)

    assert raw_sql == expected_query


@with_instances("postgres")
def test_imperative_materialize_given_config():
    _m = m2
    df = pd.DataFrame({"x": [0, 1, 2, 3]})
    tbl = Table(df)
    assert tbl.materialize() is df  # should be no-op
    assert tbl.materialize(return_nothing=True) is None

    config_context = ConfigContext.get()
    engine = config_context.store.table_store.engine
    with engine.connect() as conn:
        conn.execute(sa.text("DROP SCHEMA IF EXISTS dummy_schema CASCADE"))
        conn.execute(sa.text("CREATE SCHEMA IF NOT EXISTS dummy_schema"))
        if sa.__version__ >= "2.0.0":
            conn.commit()

    # test failure on materialization without schema
    with pytest.raises(ValueError, match="schema must be provided"):
        tbl.materialize(config_context)
    # test materialization with explicit config_context and direct schema
    ref = tbl.materialize(config_context, schema=Schema("dummy_schema"))
    sa_tbl = sa.Table(
        ref.original.name, sa.MetaData(), schema="dummy_schema", autoload_with=engine
    )
    assert [c.name for c in sa_tbl.columns] == ["x"]
    sa_tbl2 = _m.noop_lazy(sa_tbl.alias("alias"))
    assert sa_tbl2.name == "alias"
    assert sa_tbl2.original.name == sa_tbl.name

    # test return nothing materialization
    tbl.name = "a"
    assert (
        tbl.materialize(
            config_context, schema=Schema("dummy_schema"), return_nothing=True
        )
        is None
    )
    sa_tbl = sa.Table(
        tbl.name, sa.MetaData(), schema="dummy_schema", autoload_with=engine
    )
    assert [c.name for c in sa_tbl.columns] == ["x"]
    assert sa_tbl.name == tbl.name

    # test materialization with explicit config_context and prefixed schema
    tbl.name = "b"
    tbl.materialize(config_context, schema=Schema("schema", prefix="dummy_"))
    sa_tbl = sa.Table(
        tbl.name, sa.MetaData(), schema="dummy_schema", autoload_with=engine
    )
    assert [c.name for c in sa_tbl.columns] == ["x"]
    assert sa_tbl.name == tbl.name

    # test materialization with explicit config_context and postfixed schema
    tbl.name = "c"
    tbl.materialize(config_context, schema=Schema("dummy", suffix="_schema"))
    sa_tbl = sa.Table(
        tbl.name, sa.MetaData(), schema="dummy_schema", autoload_with=engine
    )
    assert [c.name for c in sa_tbl.columns] == ["x"]
    assert sa_tbl.name == tbl.name

    # test materialization with explicit config_context incl. prefixed schema
    pipedag_config = create_basic_pipedag_config(
        config_context.store.table_store.engine.url.render_as_string(
            hide_password=False
        )
    )
    raw_config2 = copy.deepcopy(pipedag_config.raw_config)
    raw_config2["instances"]["__any__"]["table_store"]["args"][
        "schema_prefix"
    ] = "dummy_"
    cfg2 = PipedagConfig(raw_config2).get()
    tbl.name = "e"
    with pytest.raises(ValueError, match="Schema prefix and postfix must match"):
        tbl.materialize(cfg2, schema=Schema("schema", prefix="dummyX_"))
    tbl.materialize(cfg2, schema=Schema("schema", prefix="dummy_"))
    sa_tbl = sa.Table(
        tbl.name, sa.MetaData(), schema="dummy_schema", autoload_with=engine
    )
    assert [c.name for c in sa_tbl.columns] == ["x"]
    assert sa_tbl.name == tbl.name

    # test materialization with explicit config_context and postfixed schema
    pipedag_config = create_basic_pipedag_config(
        config_context.store.table_store.engine.url.render_as_string(
            hide_password=False
        )
    )
    raw_config2 = copy.deepcopy(pipedag_config.raw_config)
    raw_config2["instances"]["__any__"]["table_store"]["args"][
        "schema_suffix"
    ] = "_schema"
    cfg2 = PipedagConfig(raw_config2).get()
    tbl.name = "e"
    with pytest.raises(ValueError, match="Schema prefix and postfix must match"):
        tbl.materialize(cfg2, schema=Schema("dummy", suffix="_Xschema"))
    tbl.materialize(cfg2, schema=Schema("dummy", suffix="_schema"))
    sa_tbl = sa.Table(
        tbl.name, sa.MetaData(), schema="dummy_schema", autoload_with=engine
    )
    assert [c.name for c in sa_tbl.columns] == ["x"]
    assert sa_tbl.name == tbl.name



from __future__ import annotations

import copy

import sqlalchemy as sa

from pydiverse.pipedag.backend import BaseTableStore
from pydiverse.pipedag.context import ConfigContext


def select_as(value, as_):
    return sa.select(sa.literal(value).label(as_))


def sql_table_expr(cols: dict):
    num_values = {len(vals) for vals in cols.values()}
    assert len(num_values) == 1

    queries = []
    num_values = num_values.pop()
    for i in range(num_values):
        literals = []
        for col, vals in cols.items():
            literals.append(sa.literal(vals[i]).label(col))

        queries.append(sa.select(*literals))

    return sa.union_all(*queries)


def compile_sql(query):
    engine = ConfigContext.get().store.table_store.engine
    return str(query.compile(engine, compile_kwargs={"literal_binds": True}))


def get_config_with_table_store(
    base_cfg: ConfigContext, table_store_class: type[BaseTableStore]
):
    instance = base_cfg.instance_name
    flow = base_cfg.flow_name
    cfg = ConfigContext.new(
        copy.deepcopy(base_cfg._config_dict), base_cfg.pipedag_name, flow, instance
    )
    cfg._config_dict["table_store"]["class"] = table_store_class
    # this actually instantiates the table store
    table_store = cfg.store.table_store
    assert type(table_store) == table_store_class
    return cfg



from __future__ import annotations

from pathlib import Path

import filelock
import pandas as pd
import pytest
import sqlalchemy as sa

from pydiverse.pipedag import ConfigContext, Flow, RawSql, Stage, Table, materialize
from pydiverse.pipedag.backend.table.sql.ddl import (
    CreateSchema,
    CreateTableAsSelect,
    DropTable,
)
from pydiverse.pipedag.container import ExternalTableReference, Schema
from tests.fixtures.instances import with_instances
from tests.util.sql import sql_table_expr
from tests.util.tasks_library import (
    assert_table_equal,
    noop_sql,
    simple_dataframe,
    simple_lazy_table,
)


@with_instances("ibm_db2")
def test_db2_nicknames():
    lock_path = Path(__file__).parent / "scripts" / "lock"

    @materialize(input_type=sa.Table)
    def create_nicknames(table: sa.sql.expression.Alias):
        script_path = Path(__file__).parent / "scripts" / "simple_nicknames.sql"
        simple_nicknames = Path(script_path).read_text(encoding="utf-8")
        simple_nicknames = simple_nicknames.replace(
            "{{out_schema}}", str(table.original.schema)
        )
        simple_nicknames = simple_nicknames.replace(
            "{{out_table}}", str(table.original.name)
        )

        return RawSql(simple_nicknames, "create_nicknames", separator="|")

    with Flow("f") as f:
        with Stage("stage"):
            x = simple_dataframe()
            nicknames = create_nicknames(x)
            _ = nicknames

    with filelock.FileLock(lock_path):
        # We run three times to ensure that the nicknames created in the first run
        # have to be dropped, since the same schema is reused.
        assert f.run().successful
        assert f.run().successful
        assert f.run().successful


@with_instances("ibm_db2")  # only one instance to avoid parallel DRDA WRAPPER creation
def test_db2_table_reference_nicknames():
    lock_path = Path(__file__).parent / "scripts" / "lock"

    @materialize(nout=2)
    def create_external_nicknames():
        table_store = ConfigContext.get().store.table_store
        schema = Schema("user_controlled_schema", prefix="", suffix="")
        table_name = "external_table_for_nickname"
        table_store.execute(CreateSchema(schema, if_not_exists=True))
        table_store.execute(DropTable(table_name, schema, if_exists=True))
        query = sql_table_expr({"col": [0, 1, 2, 3]})
        table_store.execute(
            CreateTableAsSelect(
                table_name,
                schema,
                query,
            )
        )
        script_path = Path(__file__).parent / "scripts" / "simple_nicknames.sql"
        simple_nicknames = Path(script_path).read_text()
        simple_nicknames = simple_nicknames.replace("{{out_schema}}", schema.get())
        simple_nicknames = simple_nicknames.replace("{{out_table}}", table_name)

        table_store.execute_raw_sql(
            RawSql(simple_nicknames, "create_external_nicknames", separator="|")
        )

        return Table(
            ExternalTableReference(
                "nick1", schema=schema.get(), shared_lock_allowed=False
            )
        ), Table(
            ExternalTableReference(
                "nick2", schema=schema.get(), shared_lock_allowed=True
            )
        )

    with Flow("f") as f:
        with Stage("stage"):
            nick_1_ref, nick_2_ref = create_external_nicknames()
            nick_1_ref_noop = noop_sql(nick_1_ref)
            nick_2_ref_noop = noop_sql(nick_2_ref)
            assert_table_equal(nick_1_ref, nick_1_ref_noop)
            assert_table_equal(nick_2_ref, nick_2_ref_noop)

    with filelock.FileLock(lock_path):
        # We run three times to ensure that the nicknames created in the first run
        # have to be dropped, since the same schema is reused.
        assert f.run().successful
        assert f.run().successful
        assert f.run().successful


@with_instances("ibm_db2_materialization_details")
@pytest.mark.parametrize("task", [simple_dataframe, simple_lazy_table])
def test_db2_table_spaces(task):
    @materialize()
    def create_table_spaces():
        script_path = Path(__file__).parent / "scripts" / "simple_table_spaces.sql"
        simple_table_spaces = Path(script_path).read_text(encoding="utf-8")
        return RawSql(simple_table_spaces, "create_table_spaces", separator="|")

    @materialize(input_type=sa.Table, lazy=False)
    def get_actual_table_space_attributes(table: sa.sql.expression.Alias):
        query = f"""
            SELECT TBSPACE, INDEX_TBSPACE, LONG_TBSPACE FROM SYSCAT.TABLES
             WHERE TABSCHEMA = '{table.original.schema.upper()}'
             AND TABNAME = '{table.original.name.upper()}'
        """
        return Table(sa.text(query), f"tbspace_attributes_{table.name}")

    @materialize(version="1.0")
    def get_expected_table_space_attributes():
        return Table(
            pd.DataFrame(
                {"TBSPACE": ["S1"], "INDEX_TBSPACE": ["S2"], "LONG_TBSPACE": ["S3"]}
            ),
            name="tbspace_attributes",
        )

    with Flow("f") as f:
        with Stage("stage", materialization_details="table_space"):
            create_table_spaces()
            x = task()
            x_attr = get_actual_table_space_attributes(x)
            attrs = get_expected_table_space_attributes()
            assert_table_equal(attrs, x_attr)

    for _ in range(3):
        assert f.run().successful



from __future__ import annotations

import pyarrow as pa

from pydiverse.pipedag.backend.table.util import DType


def test_dtype_from_pyarrow():
    def assert_conversion(type_, expected):
        assert DType.from_arrow(type_) == expected

    assert_conversion(pa.int64(), DType.INT64)
    assert_conversion(pa.int32(), DType.INT32)
    assert_conversion(pa.int16(), DType.INT16)
    assert_conversion(pa.int8(), DType.INT8)

    assert_conversion(pa.uint64(), DType.UINT64)
    assert_conversion(pa.uint32(), DType.UINT32)
    assert_conversion(pa.uint16(), DType.UINT16)
    assert_conversion(pa.uint8(), DType.UINT8)

    assert_conversion(pa.float64(), DType.FLOAT64)
    assert_conversion(pa.float32(), DType.FLOAT32)
    assert_conversion(pa.float16(), DType.FLOAT32)

    assert_conversion(pa.string(), DType.STRING)
    assert_conversion(pa.bool_(), DType.BOOLEAN)

    assert_conversion(pa.date32(), DType.DATE)
    assert_conversion(pa.date64(), DType.DATE)

    assert_conversion(pa.time32("s"), DType.TIME)
    assert_conversion(pa.time32("ms"), DType.TIME)
    assert_conversion(pa.time64("us"), DType.TIME)
    assert_conversion(pa.time64("ns"), DType.TIME)

    assert_conversion(pa.timestamp("s"), DType.DATETIME)
    assert_conversion(pa.timestamp("ms"), DType.DATETIME)
    assert_conversion(pa.timestamp("us"), DType.DATETIME)
    assert_conversion(pa.timestamp("ns"), DType.DATETIME)


def test_dtype_to_pyarrow():
    def assert_conversion(type_: DType, expected):
        assert type_.to_arrow() == expected

    assert_conversion(DType.INT64, pa.int64())
    assert_conversion(DType.INT32, pa.int32())
    assert_conversion(DType.INT16, pa.int16())
    assert_conversion(DType.INT8, pa.int8())

    assert_conversion(DType.UINT64, pa.uint64())
    assert_conversion(DType.UINT32, pa.uint32())
    assert_conversion(DType.UINT16, pa.uint16())
    assert_conversion(DType.UINT8, pa.uint8())

    assert_conversion(DType.FLOAT64, pa.float64())
    assert_conversion(DType.FLOAT32, pa.float32())

    assert_conversion(DType.STRING, pa.string())
    assert_conversion(DType.BOOLEAN, pa.bool_())

    assert_conversion(DType.DATE, pa.date32())
    assert_conversion(DType.TIME, pa.time64("us"))
    assert_conversion(DType.DATETIME, pa.timestamp("us"))
